from unittest import mock

import matplotlib.pyplot as plt
from matplotlib.testing.compare import compare_images
import numpy as np

from cabinetry.visualize import plot_result


def test_correlation_matrix(tmp_path):
    fname = tmp_path / "fig.png"
    # one parameter is below threshold so no text is shown for it on the plot
    corr_mat = np.asarray([[1.0, 0.35, 0.002], [0.35, 1.0, -0.2], [0.002, -0.2, 1.0]])
    labels = ["a", "b", "c"]

    fig = plot_result.correlation_matrix(corr_mat, labels, figure_path=fname)
    assert (
        compare_images(
            "tests/visualize/reference/correlation_matrix.png", str(fname), 0
        )
        is None
    )

    # compare figure returned by function
    fname = tmp_path / "fig_from_return.png"

    # adjust layout behavior, see https://github.com/matplotlib/matplotlib/issues/21742
    if plot_result.MPL_GEQ_36:
        fig.set_layout_engine(None)
    else:
        fig.set_constrained_layout(False)

    fig.savefig(fname)
    assert (
        compare_images(
            "tests/visualize/reference/correlation_matrix.png", str(fname), 0
        )
        is None
    )

    # do not save figure, but close it
    with mock.patch("cabinetry.visualize.utils._save_and_close") as mock_close_safe:
        fig = plot_result.correlation_matrix(corr_mat, labels, close_figure=True)
        assert mock_close_safe.call_args_list == [((fig, None, True), {})]

    plt.close("all")


def test_pulls(tmp_path):
    fname = tmp_path / "fig.png"
    bestfit = np.asarray([-0.2, 0.0, 0.1])
    uncertainty = np.asarray([0.9, 1.0, 0.7])
    labels = np.asarray(["a", "b", "c"])

    fig = plot_result.pulls(bestfit, uncertainty, labels, figure_path=fname)
    assert compare_images("tests/visualize/reference/pulls.png", str(fname), 0) is None

    # compare figure returned by function
    fname = tmp_path / "fig_from_return.png"
    fig.savefig(fname)
    assert compare_images("tests/visualize/reference/pulls.png", str(fname), 0) is None

    # do not save figure, but close it
    with mock.patch("cabinetry.visualize.utils._save_and_close") as mock_close_safe:
        fig = plot_result.pulls(bestfit, uncertainty, labels, close_figure=True)
        assert mock_close_safe.call_args_list == [((fig, None, True), {})]

    plt.close("all")


def test_ranking(tmp_path):
    fname = tmp_path / "fig.png"
    bestfit = np.asarray([0.3, -0.1])
    uncertainty = np.asarray([0.8, 1.0])
    labels = np.asarray(["jet energy scale", "modeling uncertainty"])
    impact_prefit_up = np.asarray([0.5, 0.3])
    impact_prefit_down = np.asarray([-0.4, -0.3])
    impact_postfit_up = np.asarray([0.4, 0.25])
    impact_postfit_down = np.asarray([-0.3, -0.25])

    fig = plot_result.ranking(
        bestfit,
        uncertainty,
        labels,
        impact_prefit_up,
        impact_prefit_down,
        impact_postfit_up,
        impact_postfit_down,
        figure_path=fname,
    )
    # large tolerance needed here, possibly related to lack of set_tight_layout usage
    assert (
        compare_images("tests/visualize/reference/ranking.png", str(fname), 50) is None
    )

    # compare figure returned by function
    fname = tmp_path / "fig_from_return.png"
    fig.savefig(fname)
    assert (
        compare_images("tests/visualize/reference/ranking.png", str(fname), 50) is None
    )

    # do not save figure, but close it
    with mock.patch("cabinetry.visualize.utils._save_and_close") as mock_close_safe:
        fig = plot_result.ranking(
            bestfit,
            uncertainty,
            labels,
            impact_prefit_up,
            impact_prefit_down,
            impact_postfit_up,
            impact_postfit_down,
            close_figure=True,
        )
        assert mock_close_safe.call_args_list == [((fig, None, True), {})]

    plt.close("all")


def test_scan(tmp_path):
    fname = tmp_path / "fig.png"
    par_name = "a"
    par_mle = 1.5
    par_unc = 0.2
    par_vals = np.asarray([1.1, 1.3, 1.5, 1.7, 1.9])
    par_nlls = np.asarray([4.1, 1.0, 0.0, 1.1, 3.9])

    fig = plot_result.scan(
        par_name, par_mle, par_unc, par_vals, par_nlls, figure_path=fname
    )
    assert compare_images("tests/visualize/reference/scan.png", str(fname), 0) is None

    # compare figure returned by function
    fname = tmp_path / "fig_from_return.png"
    fig.savefig(fname)
    assert compare_images("tests/visualize/reference/scan.png", str(fname), 0) is None

    # do not save figure, but close it
    with mock.patch("cabinetry.visualize.utils._save_and_close") as mock_close_safe:
        # no 68% CL / 95% CL text
        par_nlls = np.asarray([0.1, 0.04, 0.0, 0.04, 0.1])
        fig = plot_result.scan(
            par_name, par_mle, par_unc, par_vals, par_nlls, close_figure=True
        )
        assert mock_close_safe.call_args_list == [((fig, None, True), {})]

    plt.close("all")


def test_limit(tmp_path):
    fname = tmp_path / "fig.png"
    observed_CLs = np.asarray([0.31, 0.05, 0.005, 0.0001])
    expected_CLs = np.asarray(
        [
            [0.07, 0.16, 0.35, 0.62, 0.88],
            [0.0, 0.01, 0.06, 0.23, 0.56],
            [0.0, 0.0, 0.0, 0.05, 0.2],
            [0.0, 0.0, 0.0, 0.0, 0.04],
        ]
    )
    poi_values = np.asarray([0.5, 1.0, 1.5, 2.0])
    cls_target = 0.05

    fig = plot_result.limit(
        observed_CLs, expected_CLs, poi_values, cls_target, figure_path=fname
    )
    assert compare_images("tests/visualize/reference/limit.png", str(fname), 0) is None

    # compare figure returned by function
    fname = tmp_path / "fig_from_return.png"
    fig.savefig(fname)
    assert compare_images("tests/visualize/reference/limit.png", str(fname), 0) is None

    # do not save figure, but close it
    with mock.patch("cabinetry.visualize.utils._save_and_close") as mock_close_safe:
        fig = plot_result.limit(
            observed_CLs, expected_CLs, poi_values, cls_target, close_figure=True
        )
        assert mock_close_safe.call_args_list == [((fig, None, True), {})]

    plt.close("all")



"""Provides utilities for template histogram handling."""

import pathlib
from typing import Any, Dict, List, Literal, Optional, Union

from cabinetry import histo


def _check_for_override(
    systematic: Dict[str, Any], template: Literal["Up", "Down"], option: str
) -> Optional[Union[str, List[str]]]:
    """Returns an override if specified by a template of a systematic.

    Given a systematic and a string specifying which template is currently under
    consideration, check whether the systematic defines an override for an option.
    Return the override if it exists, otherwise return None.

    Args:
        systematic (Dict[str, Any]): containing all systematic information
        template (Literal["Up", "Down"]): template considered: "Up" or "Down"
        option (str): the option for which the presence of an override is checked

    Returns:
        Optional[Union[str, List[str]]]: either None if no override exists, or the
        override
    """
    return systematic.get(template, {}).get(option, None)


def _name_and_save(
    histogram_folder: pathlib.Path,
    histogram: histo.Histogram,
    region: Dict[str, Any],
    sample: Dict[str, Any],
    systematic: Dict[str, Any],
    template: Optional[Literal["Up", "Down"]],
) -> None:
    """Generates a unique name for a histogram and saves the histogram.

    Args:
        histogram_folder (pathlib.Path): folder to save the histograms to
        histogram (histo.Histogram): histogram to save
        region (Dict[str, Any]): containing all region information
        sample (Dict[str, Any]): containing all sample information
        systematic (Dict[str, Any]): containing all systematic
        template (Optional[Literal["Up", "Down"]]): template considered: "Up",
            "Down", or None for nominal
    """
    # generate a name for the histogram
    histogram_name = histo.name(region, sample, systematic, template=template)

    # check the histogram for common issues
    histogram.validate(histogram_name)

    # save it
    histo_path = histogram_folder / histogram_name
    histogram.save(histo_path)



import logging
from unittest import mock

import jsonschema
import pytest

from cabinetry import configuration


@mock.patch("cabinetry.configuration.validate")
def test_load(mock_validation):
    conf = configuration.load("config_example.yml")
    assert isinstance(conf, dict)
    mock_validation.assert_called_once()


def test_validate():
    config_valid = {
        "General": {"Measurement": "", "HistogramFolder": "", "InputPath": ""},
        "Regions": [{"Name": "", "Filter": "", "Variable": "", "Binning": [0, 1]}],
        "Samples": [{"Name": "", "Tree": "", "Data": True}],
        "NormFactors": [{"Name": ""}],
    }
    assert configuration.validate(config_valid)

    # not exactly one data sample
    config_multiple_data_samples = {
        "General": {"Measurement": "", "HistogramFolder": "", "InputPath": ""},
        "Regions": [{"Name": "", "Filter": "", "Variable": "", "Binning": [0, 1]}],
        "Samples": [{"Name": "", "Tree": ""}],
        "NormFactors": [{"Name": ""}],
    }
    with pytest.raises(
        NotImplementedError, match="can only handle cases with exactly one data sample"
    ):
        configuration.validate(config_multiple_data_samples)

    # config doesn't adhere to schema
    config_schema_mismatch = {}
    with pytest.raises(
        jsonschema.exceptions.ValidationError, match="'General' is a required property"
    ):
        configuration.validate(config_schema_mismatch)

    # schema cannot be loaded
    with mock.patch(
        "cabinetry.configuration.pkgutil.get_data", return_value=None
    ) as mock_get:
        with pytest.raises(FileNotFoundError, match="could not load config schema"):
            configuration.validate({})
        mock_get.assert_called_once()

    # uniqueness of names
    with mock.patch("jsonschema.validate", return_value=None):
        config_region_names = {
            "Regions": [{"Name": "abc"}, {"Name": "abc"}],
            "Samples": [{"Name": "", "Tree": "", "Data": True}],
            "NormFactors": [],
        }
        with pytest.raises(ValueError, match="all region names must be unique"):
            configuration.validate(config_region_names)

        config_sample_names = {
            "Regions": [],
            "Samples": [{"Name": "abc", "Tree": "", "Data": True}, {"Name": "abc"}],
            "NormFactors": [],
        }
        with pytest.raises(ValueError, match="all sample names must be unique"):
            configuration.validate(config_sample_names)

        config_normfactor_names = {
            "Regions": [],
            "Samples": [{"Name": "", "Tree": "", "Data": True}],
            "NormFactors": [{"Name": "abc"}, {"Name": "abc"}],
        }
        with pytest.raises(ValueError, match="all normfactor names must be unique"):
            configuration.validate(config_normfactor_names)

        config_systematic_names = {
            "Regions": [],
            "Samples": [{"Name": "abc", "Tree": "", "Data": True}],
            "Systematics": [{"Name": "abc"}, {"Name": "abc"}],
            "NormFactors": [],
        }
        with pytest.raises(ValueError, match="all systematic names must be unique"):
            configuration.validate(config_systematic_names)


def test_print_overview(caplog):
    caplog.set_level(logging.DEBUG)
    config_example = {
        "General": "",
        "Regions": "",
        "NormFactors": "",
        "Samples": [{"Data": True}],
        "Systematics": "",
    }
    configuration.print_overview(config_example)
    assert "  1 Sample(s)" in [rec.message for rec in caplog.records]
    assert "  0 Regions(s)" in [rec.message for rec in caplog.records]
    assert "  0 NormFactor(s)" in [rec.message for rec in caplog.records]
    assert "  0 Systematic(s)" in [rec.message for rec in caplog.records]
    caplog.clear()

    config_example_no_sys = {
        "General": "",
        "Regions": "",
        "NormFactors": "",
        "Samples": [{"Data": True}],
    }
    configuration.print_overview(config_example_no_sys)
    assert "Systematic(s)" not in [rec.message for rec in caplog.records]
    caplog.clear()


@pytest.mark.parametrize(
    "samples, converted", [("sample", ["sample"]), (["sample"], ["sample"])]
)
def test__setting_to_list(samples, converted):
    assert configuration._setting_to_list(samples) == converted


@pytest.mark.parametrize(
    "x_y_key, contained",
    [
        (({"Name": "abc"}, {}, "key"), True),
        (({"Name": "abc"}, {"key": ["abc", "def"]}, "key"), True),
        (({"Name": "abc"}, {"key": ["def"]}, "key"), False),
        (({"Name": "abc"}, {"key": "abc"}, "key"), True),
        (({"Name": "abc"}, {"key": "def"}, "key"), False),
    ],
)
def test__x_contains_y(x_y_key, contained):
    assert configuration._x_contains_y(*x_y_key) is contained


@pytest.mark.parametrize(
    "region_and_sample, contained",
    [
        (({"Name": "CR"}, {}), True),
        (({"Name": "SR"}, {"Regions": ["SR", "CR"]}), True),
        (({"Name": "CR"}, {"Regions": "SR"}), False),
    ],
)
def test_region_contains_sample(region_and_sample, contained):
    assert configuration.region_contains_sample(*region_and_sample) is contained


@pytest.mark.parametrize(
    "region_and_modifier, contained",
    [
        (({"Name": "CR"}, {}), True),
        (({"Name": "SR"}, {"Regions": ["SR", "CR"]}), True),
        (({"Name": "CR"}, {"Regions": "SR"}), False),
    ],
)
def test_region_contains_modifier(region_and_modifier, contained):
    assert configuration.region_contains_modifier(*region_and_modifier) is contained


@pytest.mark.parametrize(
    "sample_and_modifier, contained",
    [
        (({"Name": "Signal"}, {}), True),
        (({"Name": "Signal"}, {"Samples": ["Signal", "Background"]}), True),
        (({"Name": "Signal"}, {"Samples": "Background"}), False),
    ],
)
def test_sample_contains_modifier(sample_and_modifier, contained):
    assert configuration.sample_contains_modifier(*sample_and_modifier) is contained


@pytest.mark.parametrize(
    "reg_sam_sys_tem, is_needed",
    [
        # nominal
        (({}, {}, {"Name": "abc"}, None), True),
        # non-nominal data
        (({}, {"Data": True}, {"Name": "var"}, "Up"), False),
        # overall normalization variation
        (({}, {}, {"Type": "Normalization"}, "Up"), False),
        # normalization + shape variation
        (({}, {"Name": "Signal"}, {"Type": "NormPlusShape"}, "Up"), True),
        # normalization + shape variation on specified and affected sample
        (
            (
                {},
                {"Name": "Signal"},
                {"Type": "NormPlusShape", "Samples": ["Signal", "Background"]},
                "Up",
            ),
            True,
        ),
        # normalization + shape variation on non-affected sample
        (
            (
                {},
                {"Name": "Background"},
                {"Type": "NormPlusShape", "Samples": "Signal"},
                "Up",
            ),
            False,
        ),
        # template not needed due to symmetrization
        (
            (
                {},
                {"Name": "Signal"},
                {"Type": "NormPlusShape", "Up": {"Symmetrize": True}},
                "Up",
            ),
            False,
        ),
        # template needed since symmetrization is only for other direction
        (
            (
                {},
                {"Name": "Signal"},
                {
                    "Type": "NormPlusShape",
                    "Samples": "Signal",
                    "Down": {"Symmetrize": True},
                },
                "Up",
            ),
            True,
        ),
        # region does not contain sample
        (
            (
                {"Name": "CR"},
                {"Name": "Signal", "Regions": "SR"},
                {"Type": "NormPlusShape"},
                None,
            ),
            False,
        ),
        # region does contain sample
        (
            (
                {"Name": "CR"},
                {"Name": "Signal", "Regions": "CR"},
                {"Type": "NormPlusShape"},
                None,
            ),
            True,
        ),
        # region does not contain modifier
        (
            (
                {"Name": "CR"},
                {"Name": "Signal"},
                {"Type": "NormPlusShape", "Regions": "SR"},
                "Up",
            ),
            False,
        ),
        # region does contain modifier
        (
            (
                {"Name": "CR"},
                {"Name": "Signal"},
                {"Type": "NormPlusShape", "Regions": "CR"},
                "Up",
            ),
            True,
        ),
    ],
)
def test_histogram_is_needed(reg_sam_sys_tem, is_needed):
    # could also mock region_contains_sample, region_contains_modifier, and
    # sample_contains_modifier
    reg, sam, sys, tem = reg_sam_sys_tem
    assert configuration.histogram_is_needed(*reg_sam_sys_tem) is is_needed


def test_histogram_is_needed_unknown():
    # non-supported systematic
    with pytest.raises(ValueError, match="unknown systematics type: unknown"):
        configuration.histogram_is_needed({}, {}, {"Type": "unknown"}, "Up")


def test_region_dict(caplog):
    caplog.set_level(logging.WARNING)

    config = {"Regions": [{"Name": "reg_a"}, {"Name": "reg_b"}]}
    assert configuration.region_dict(config, "reg_a") == {"Name": "reg_a"}

    config = {"Regions": [{"Name": "reg_a"}, {"Name": "reg_a"}]}
    assert configuration.region_dict(config, "reg_a") == {"Name": "reg_a"}
    assert "found more than one region with name reg_a" in [
        rec.message for rec in caplog.records
    ]

    with pytest.raises(ValueError, match="region abc not found in config"):
        configuration.region_dict(config, "abc")



import numpy as np
import pyhf
import pytest

import cabinetry


@pytest.fixture
def reset_backend():
    """This sets the ``pyhf`` backend back to ``numpy`` after a test has run."""
    # setup before test
    yield  # test
    # teardown after test
    pyhf.set_backend("numpy")


@pytest.mark.slow
@pytest.mark.no_cover
@pytest.mark.parametrize("backend", ["jax", "pytorch", "tensorflow"])
def test_backend_integration(backend, reset_backend):
    """Integration test for the inference pipeline that can be run with all ``pyhf``
    backends to ensure they work. ``typeguard`` will catch type issues at runtime.
    """
    pyhf.set_backend(backend)
    # construct a simple workspace
    model = pyhf.simplemodels.uncorrelated_background(
        signal=[24.0, 22.0], bkg=[50.0, 52.0], bkg_uncertainty=[3.0, 7.0]
    )
    ws = pyhf.Workspace.build(model, [74, 74])

    model, data = cabinetry.model_utils.model_and_data(ws)

    # MLE comparison of both implementations
    fit_results_1 = cabinetry.fit.fit(model, data, minos="mu", goodness_of_fit=True)
    fit_results_2 = cabinetry.fit.fit(
        model, data, minos="mu", goodness_of_fit=True, custom_fit=True
    )
    assert np.allclose(fit_results_1.bestfit, fit_results_2.bestfit, rtol=0.01)
    assert np.allclose(fit_results_1.uncertainty, fit_results_2.uncertainty, rtol=0.01)
    assert np.allclose(fit_results_1.corr_mat, fit_results_2.corr_mat, rtol=0.01)
    assert np.allclose(fit_results_1.best_twice_nll, fit_results_2.best_twice_nll)
    assert np.allclose(fit_results_1.goodness_of_fit, fit_results_2.goodness_of_fit)

    # remaining types of fit results
    cabinetry.fit.ranking(model, data)
    cabinetry.fit.scan(model, data, "mu")
    cabinetry.fit.limit(model, data)
    cabinetry.fit.significance(model, data)

    # model_utils functions that deal with expected_data
    cabinetry.model_utils.asimov_data(model)
    # parameters, uncertainty, correlation for stdev
    param_values = cabinetry.model_utils.asimov_parameters(model)
    param_uncertainty = cabinetry.model_utils.prefit_uncertainties(model)
    corr_mat = np.zeros(shape=(len(param_values), len(param_values)))
    np.fill_diagonal(corr_mat, 1.0)
    cabinetry.model_utils.yield_stdev(model, param_values, param_uncertainty, corr_mat)


def test_backend_reset():
    # ensure that the pyhf backend is numpy again
    assert pyhf.tensorlib.name == "numpy"



"""Provides utilities to handle the cabinetry configuration."""

import json
import logging
import pathlib
import pkgutil
from typing import Any, Dict, List, Literal, Optional, Union

import jsonschema
import yaml


log = logging.getLogger(__name__)


def load(file_path_string: Union[str, pathlib.Path]) -> Dict[str, Any]:
    """Loads, validates, and returns a config file from the provided path.

    Args:
        file_path_string (Union[str, pathlib.Path]): path to config file

    Returns:
        Dict[str, Any]: cabinetry configuration
    """
    file_path = pathlib.Path(file_path_string)
    log.info(f"opening config file {file_path}")
    config = yaml.safe_load(file_path.read_text())
    validate(config)
    return config


def validate(config: Dict[str, Any]) -> bool:
    """Returns True if the config file is validated, otherwise raises exceptions.

    Checks that the config satisfies the json schema, and performs additional checks to
    validate the config further.

    Args:
        config (Dict[str, Any]): cabinetry configuration

    Raises:
        NotImplementedError: when more than one data sample is found
        ValueError: when region / sample / normfactor / systematic names are not unique

    Returns:
        bool: whether the validation was successful
    """
    # load json schema for config and validate against it
    schema_text = pkgutil.get_data(__name__, "schemas/config.json")
    if schema_text is None:
        raise FileNotFoundError("could not load config schema")
    config_schema = json.loads(schema_text)
    jsonschema.validate(instance=config, schema=config_schema)

    # check that there is exactly one data sample
    if sum(sample.get("Data", False) for sample in config["Samples"]) != 1:
        raise NotImplementedError("can only handle cases with exactly one data sample")

    # check that region names are unique
    region_names = [region["Name"] for region in config["Regions"]]
    if len(set(region_names)) != len(region_names):
        raise ValueError(f"all region names must be unique: {region_names}")

    # check that sample names are unique
    sample_names = [sample["Name"] for sample in config["Samples"]]
    if len(set(sample_names)) != len(sample_names):
        raise ValueError(f"all sample names must be unique: {sample_names}")

    # check that normfactor names are unique
    # may technically not be required, but non-unique names do not seem to offer any
    # obvious advantages, so require uniqueness here for consistency
    normfactor_names = [normfactor["Name"] for normfactor in config["NormFactors"]]
    if len(set(normfactor_names)) != len(normfactor_names):
        raise ValueError(f"all normfactor names must be unique: {normfactor_names}")

    # check that systematic names are unique
    # systematics are optional, may be empty
    systematic_names = [sys["Name"] for sys in config.get("Systematics", [])]
    if len(set(systematic_names)) != len(systematic_names):
        raise ValueError(f"all systematic names must be unique: {systematic_names}")

    # if no issues are found
    return True


def print_overview(config: Dict[str, Any]) -> None:
    """Prints a compact summary of a config file.

    Args:
        config (Dict[str, Any]): cabinetry configuration
    """
    log.info("the config contains:")
    log.info(f"  {len(config['Samples'])} Sample(s)")
    log.info(f"  {len(config['Regions'])} Regions(s)")
    log.info(f"  {len(config['NormFactors'])} NormFactor(s)")
    if "Systematics" in config.keys():
        log.info(f"  {len(config['Systematics'])} Systematic(s)")


def _setting_to_list(setting: Union[str, List[str]]) -> List[str]:
    """Converts a configuration setting to a list.

    The config allows for two ways of specifying some settings, for example samples. A
    single sample is specified as ``"Samples": "ABC"``, a list of samples as
    ``"Samples": ["ABC", "DEF"]``. For consistent treatment, the single sample is
    converted to a list.

    Args:
        setting (Union[str, List[str]]): name of single setting value or list of values

    Returns:
        list: name(s) of sample(s)
    """
    if not isinstance(setting, list):
        setting = [setting]
    return setting


def _x_contains_y(x: Dict[str, Any], y: Dict[str, Any], y_key: str) -> bool:
    """Checks if object ``x`` contains ``y`` using property ``y_key`` of ``y``.

    If ``y_key`` is not specified, ``x`` is assumed to contain ``y`` by default. Used
    to check if regions contain samples/modifiers and if samples contain modifiers.
    ``x`` is identified by its "Name" property, which must exist.

    Args:
        x (Dict[str, Any]): containing all relevant information: region or sample, must
            have a "Name" property
        y (Dict[str, Any]): containing all relevant information: sample or modifier
        y_key (str): property of ``y`` to check

    Returns:
        bool: True if ``x`` contains ``y``, False otherwise
    """
    # y_key setting of y is optional, default to empty list
    matched_x_list = _setting_to_list(y.get(y_key, []))
    if matched_x_list and x["Name"] not in matched_x_list:
        # only some x are allowed as specified in list, and current x does not match
        return False
    return True


def region_contains_sample(region: Dict[str, Any], sample: Dict[str, Any]) -> bool:
    """Checks if a region contains a given sample.

    A sample enters all regions by default, and its "Regions" property can be used to
    specify a single region or list of regions that contain the sample.

    Args:
        region (Dict[str, Any]): containing all region information
        sample (Dict[str, Any]): containing all sample information

    Returns:
        bool: True if region contains sample, False otherwise
    """
    return _x_contains_y(region, sample, "Regions")


def region_contains_modifier(region: Dict[str, Any], modifier: Dict[str, Any]) -> bool:
    """Checks if a region contains a given modifier (Systematic, NormFactor).

    A modifier affects all regions by default, and its "Regions" property can be used to
    specify a single region or list of regions that contain the modifier. This does not
    check whether the modifier only acts on samples which the region does not contain.

    Args:
        region (Dict[str, Any]): containing all region information
        modifier (Dict[str, Any]): containing all modifier information (a Systematic or
            a NormFactor)

    Returns:
        bool: True if region contains modifier, False otherwise
    """
    return _x_contains_y(region, modifier, "Regions")


def sample_contains_modifier(sample: Dict[str, Any], modifier: Dict[str, Any]) -> bool:
    """Checks if a sample is affected by a given modifier (Systematic, NormFactor).

    A modifier affects all samples by default, and its "Samples" property can be used to
    specify a single sample or list of samples on which the modifier acts.

    Args:
        sample (Dict[str, Any]): containing all sample information
        modifier (Dict[str, Any]): containing all modifier information (a Systematic or
            a NormFactor)

    Returns:
        bool: True if sample is affected, False otherwise
    """
    return _x_contains_y(sample, modifier, "Samples")


def histogram_is_needed(
    region: Dict[str, Any],
    sample: Dict[str, Any],
    systematic: Dict[str, Any],
    template: Optional[Literal["Up", "Down"]],
) -> bool:
    """Determines whether a histogram is needed for a specific configuration.

    The configuration is defined by the region, sample, systematic and template ("Up" or
    "Down", None for nominal).

    Args:
        region (Dict[str, Any]): containing all region information
        sample (Dict[str, Any]): containing all sample information
        systematic (Dict[str, Any]): containing all systematic information
        template (Optional[Literal["Up", "Down"]]): which template to consider: "Up",
            "Down", None for the nominal case

    Raises:
        NotImplementedError: non-supported systematic variations based on histograms are
            requested

    Returns:
        bool: whether a histogram is needed
    """
    if not region_contains_sample(region, sample):
        # region does not contain sample
        return False

    if template is None:
        # for nominal, histograms are generally needed
        histo_needed = True
    else:
        # non-nominal case
        if sample.get("Data", False):
            # for data, only nominal histograms are needed
            histo_needed = False
        else:
            # handle non-nominal, non-data histograms
            # this assumes that the systematic dict satisfies config schema requirements
            if systematic["Type"] == "Normalization":
                # no histogram needed for normalization variation
                histo_needed = False
            elif systematic["Type"] == "NormPlusShape":
                # for a variation defined via a template, a histogram is needed (if
                # sample is affected in region)
                histo_needed = region_contains_modifier(region, systematic)
                histo_needed &= sample_contains_modifier(sample, systematic)
                # if symmetrization is specified for the template under consideration,
                # a histogram is not needed (since it will later on be obtained via
                # symmetrization)
                if systematic.get(template, {}).get("Symmetrize", False):
                    histo_needed = False
            else:
                raise ValueError(f"unknown systematics type: {systematic['Type']}")

    return histo_needed


def region_dict(config: Dict[str, Any], region_name: str) -> Dict[str, Any]:
    """Returns the dictionary for a region with the given name.

    Args:
        config (Dict[str, Any]): cabinetry configuration file
        region_name (str): name of region

    Raises:
        ValueError: when region is not found in config

    Returns:
        Dict[str, Any]: dictionary describing region
    """
    regions = [reg for reg in config["Regions"] if reg["Name"] == region_name]
    if len(regions) == 0:
        raise ValueError(f"region {region_name} not found in config")
    if len(regions) > 1:
        log.error(f"found more than one region with name {region_name}")
    return regions[0]



import copy
import logging
from unittest import mock

import numpy as np
import pyhf
import pytest

from cabinetry import model_utils
from cabinetry.fit.results_containers import FitResults


def test_ModelPrediction(example_spec):
    model = pyhf.Workspace(example_spec).model()
    model_yields = [[[10.0]]]
    total_stdev_model_bins = [[[2.0], [2.0]]]
    total_stdev_model_channels = [[2.0, 2.0]]
    label = "abc"
    model_prediction = model_utils.ModelPrediction(
        model, model_yields, total_stdev_model_bins, total_stdev_model_channels, label
    )
    assert model_prediction.model == model
    assert model_prediction.model_yields == model_yields
    assert model_prediction.total_stdev_model_bins == total_stdev_model_bins
    assert model_prediction.total_stdev_model_channels == total_stdev_model_channels
    assert model_prediction.label == label


def test_model_and_data(example_spec):
    model, data = model_utils.model_and_data(example_spec)
    assert model.spec["channels"] == example_spec["channels"]
    assert model.config.modifier_settings == {
        "normsys": {"interpcode": "code4"},
        "histosys": {"interpcode": "code4p"},
    }
    assert data == [475, 1.0]

    # requesting Asimov dataset
    model, data = model_utils.model_and_data(example_spec, asimov=True)
    assert data == [51.8, 1.0]

    # without auxdata
    model, data = model_utils.model_and_data(example_spec, include_auxdata=False)
    assert data == [475]


def test_asimov_data(example_spec):
    model = pyhf.Workspace(example_spec).model()
    assert model_utils.asimov_data(model) == [51.8, 1]

    # without auxdata
    assert model_utils.asimov_data(model, include_auxdata=False) == [51.8]

    # respect nominal settings for normfactors
    example_spec["measurements"][0]["config"]["parameters"].append(
        {"name": "Signal strength", "inits": [2.0]}
    )
    model = pyhf.Workspace(example_spec).model()
    assert model_utils.asimov_data(model, include_auxdata=False) == [103.6]

    # post-fit Asimov
    fit_results = FitResults(
        np.asarray([3.0, 1.0]),
        np.asarray([0.1, 0.01]),
        ["Signal strength", "staterror_Signal-Region[0]"],
        np.asarray([[1.0, 0.1], [0.1, 1.0]]),
        0.0,
    )
    assert np.allclose(
        model_utils.asimov_data(model, fit_results=fit_results), [155.4, 1.0]
    )

    # post-fit Asimov, custom POI value (name from model config)
    assert np.allclose(
        model_utils.asimov_data(model, fit_results=fit_results, poi_value=1.5),
        [77.7, 1.0],
    )

    # pre-fit Asimov, custom POI name + value
    assert np.allclose(
        model_utils.asimov_data(
            model, poi_name="staterror_Signal-Region[0]", poi_value=1.1
        ),
        [113.96, 1.1],  # 2*51.8*1.1
    )

    # post-fit Asimov, no POI in model and no poi_name either
    example_spec["measurements"][0]["config"]["poi"] = ""
    model = pyhf.Workspace(example_spec).model()
    with pytest.raises(
        ValueError,
        match="no POI specified in model, use the poi_name argument to set POI name",
    ):
        model_utils.asimov_data(model, poi_value=1.0)


def test_asimov_parameters(example_spec, example_spec_shapefactor, example_spec_lumi):
    model = pyhf.Workspace(example_spec).model()
    pars = model_utils.asimov_parameters(model)
    assert np.allclose(pars, [1.0, 1.0])

    # respect shapefactor initial values
    example_spec_shapefactor["measurements"][0]["config"]["parameters"].append(
        {"name": "shape factor", "inits": [1.2, 1.1]}
    )
    model = pyhf.Workspace(example_spec_shapefactor).model()
    pars = model_utils.asimov_parameters(model)
    assert np.allclose(pars, [1.0, 1.2, 1.1])

    # respect normfactor initial values
    normfactor_spec = copy.deepcopy(example_spec)
    normfactor_spec["measurements"][0]["config"]["parameters"].append(
        {"name": "Signal strength", "inits": [2.0]}
    )
    model = pyhf.Workspace(normfactor_spec).model()
    pars = model_utils.asimov_parameters(model)
    assert np.allclose(pars, [2.0, 1.0])

    # modifier with nominal value 0 and different initial value (which is ignored)
    normsys_spec = copy.deepcopy(example_spec)
    normsys_spec["channels"][0]["samples"][0]["modifiers"].append(
        {"data": {"hi": 1.2, "lo": 0.8}, "name": "normsys_example", "type": "normsys"}
    )
    normsys_spec["measurements"][0]["config"]["parameters"].append(
        {"name": "normsys_example", "inits": [0.5]}
    )
    model = pyhf.Workspace(normsys_spec).model()
    pars = model_utils.asimov_parameters(model)
    assert np.allclose(pars, [1.0, 0.0, 1.0])

    # shapesys modifier with nominal value 1 and different initial value (ignored)
    shapesys_spec = copy.deepcopy(example_spec)
    shapesys_spec["channels"][0]["samples"][0]["modifiers"].append(
        {"data": [5.0], "name": "shapesys_example", "type": "shapesys"}
    )
    shapesys_spec["measurements"][0]["config"]["parameters"].append(
        {"name": "shapesys_example", "inits": [1.5]}
    )
    model = pyhf.Workspace(shapesys_spec).model()
    pars = model_utils.asimov_parameters(model)
    assert np.allclose(pars, [1.0, 1.0, 1.0])

    # lumi modifier with nominal value 1 and different initial value (ignored)
    model = pyhf.Workspace(example_spec_lumi).model()
    pars = model_utils.asimov_parameters(model)
    assert np.allclose(pars, [1.0, 1.0])


def test_prefit_uncertainties(
    example_spec,
    example_spec_multibin,
    example_spec_shapefactor,
    example_spec_zero_staterror,
):
    model = pyhf.Workspace(example_spec).model()
    unc = model_utils.prefit_uncertainties(model)
    assert np.allclose(unc, [0.0, 0.0])  # fixed parameter and normfactor

    model = pyhf.Workspace(example_spec_multibin).model()
    unc = model_utils.prefit_uncertainties(model)
    assert np.allclose(unc, [0.0, 0.2, 0.4, 0.125])

    model = pyhf.Workspace(example_spec_shapefactor).model()
    unc = model_utils.prefit_uncertainties(model)
    assert np.allclose(unc, [0.0, 0.0, 0.0])

    model = pyhf.Workspace(example_spec_zero_staterror).model()
    unc = model_utils.prefit_uncertainties(model)
    assert np.allclose(unc, [0.0, 0.2, 0.0])  # partially fixed staterror


def test__hashable_model_key(example_spec):
    # key matches for two models built from the same spec
    model_1 = pyhf.Workspace(example_spec).model()
    model_2 = pyhf.Workspace(example_spec).model()
    assert model_utils._hashable_model_key(model_1) == model_utils._hashable_model_key(
        model_2
    )

    # key does not match if model has different interpcode
    model_new_interpcode = pyhf.Workspace(example_spec).model(
        modifier_settings={
            "normsys": {"interpcode": "code1"},
            "histosys": {"interpcode": "code0"},
        }
    )
    assert model_utils._hashable_model_key(model_1) != model_utils._hashable_model_key(
        model_new_interpcode
    )


def test_yield_stdev(example_spec, example_spec_multibin):
    model = pyhf.Workspace(example_spec).model()
    parameters = np.asarray([0.95, 1.05])
    uncertainty = np.asarray([0.1, 0.1])
    corr_mat = np.asarray([[1.0, 0.2], [0.2, 1.0]])

    total_stdev_bin, total_stdev_chan = model_utils.yield_stdev(
        model, parameters, uncertainty, corr_mat
    )
    assert np.allclose(total_stdev_bin, [[[8.03150606], [8.03150606]]])
    assert np.allclose(total_stdev_chan, [[8.03150606, 8.03150606]])

    # pre-fit
    parameters = np.asarray([1.0, 1.0])
    uncertainty = np.asarray([0.0, 0.0495665682])
    diag_corr_mat = np.diagflat([1.0, 1.0])
    total_stdev_bin, total_stdev_chan = model_utils.yield_stdev(
        model, parameters, uncertainty, diag_corr_mat
    )
    assert np.allclose(total_stdev_bin, [[[2.56754823], [2.56754823]]])  # the staterror
    assert np.allclose(total_stdev_chan, [[2.56754823, 2.56754823]])

    # multiple channels, bins, staterrors
    model = pyhf.Workspace(example_spec_multibin).model()
    parameters = np.asarray([1.3, 0.9, 1.05, 0.95])
    uncertainty = np.asarray([0.3, 0.1, 0.05, 0.1])
    corr_mat = np.asarray(
        [
            [1.0, 0.2, 0.2, 0.3],
            [0.2, 1.0, 0.1, 0.1],
            [0.2, 0.1, 1.0, 0.3],
            [0.3, 0.1, 0.3, 1.0],
        ]
    )
    total_stdev_bin, total_stdev_chan = model_utils.yield_stdev(
        model, parameters, uncertainty, corr_mat
    )
    expected_stdev_bin = [
        [[8.056054, 1.670629], [8.056054, 1.670629]],
        [[2.775377], [2.775377]],
    ]
    expected_stdev_chan = [[9.596340, 9.596340], [2.775377, 2.775377]]
    for i_reg in range(2):  # two channels
        assert np.allclose(total_stdev_bin[i_reg], expected_stdev_bin[i_reg])
        assert np.allclose(total_stdev_chan[i_reg], expected_stdev_chan[i_reg])

    # test caching by calling again with same arguments
    total_stdev_bin, total_stdev_chan = model_utils.yield_stdev(
        model, parameters, uncertainty, corr_mat
    )
    for i_reg in range(2):
        assert np.allclose(total_stdev_bin[i_reg], expected_stdev_bin[i_reg])
        assert np.allclose(total_stdev_chan[i_reg], expected_stdev_chan[i_reg])
    # also look up cache directly
    from_cache = model_utils._YIELD_STDEV_CACHE[
        model_utils._hashable_model_key(model),
        tuple(parameters),
        tuple(uncertainty),
        corr_mat.tobytes(),
    ]
    for i_reg in range(2):
        assert np.allclose(from_cache[0][i_reg], expected_stdev_bin[i_reg])
        assert np.allclose(from_cache[1][i_reg], expected_stdev_chan[i_reg])


@mock.patch(
    "cabinetry.model_utils.yield_stdev",
    side_effect=[
        (
            [[[5.0, 2.0], [5.0, 2.0]], [[1.0], [1.0]]],
            [[5.38516481, 5.38516481], [1.0, 1.0]],
        ),
        ([[[0.3], [0.3]]], [[0.3, 0.3]]),
        ([[[0.3], [0.3]]], [[0.3, 0.3]]),
    ],
)
@mock.patch(
    "cabinetry.model_utils.prefit_uncertainties",
    return_value=np.asarray([0.0, 0.2, 0.4, 0.125]),
)
@mock.patch(
    "cabinetry.model_utils.asimov_parameters",
    side_effect=[
        np.asarray([1.0, 1.0, 1.0, 1.0]),
        np.asarray([1.0, 1.0]),
        np.asarray([1.0, 1.0]),
    ],
)
def test_prediction(
    mock_asimov, mock_unc, mock_stdev, example_spec_multibin, example_spec, caplog
):
    caplog.set_level(logging.DEBUG)
    model = pyhf.Workspace(example_spec_multibin).model()

    # pre-fit prediction, multi-channel model
    model_pred = model_utils.prediction(model)
    assert model_pred.model == model
    # yields from pyhf expected_data call, per-bin / per-channel uncertainty from mock
    assert model_pred.model_yields == [[[25.0, 5.0]], [[8.0]]]
    assert model_pred.total_stdev_model_bins == [
        [[5.0, 2.0], [5.0, 2.0]],
        [[1.0], [1.0]],
    ]
    assert np.allclose(
        model_pred.total_stdev_model_channels, [[5.38516481, 5.38516481], [1.0, 1.0]]
    )
    assert model_pred.label == "pre-fit"

    # Asimov parameter calculation and pre-fit uncertainties
    assert mock_asimov.call_args_list == [((model,), {})]
    assert mock_unc.call_args_list == [((model,), {})]

    # call to stdev calculation
    assert mock_stdev.call_count == 1
    assert mock_stdev.call_args_list[0][0][0] == model
    assert np.allclose(mock_stdev.call_args_list[0][0][1], [1.0, 1.0, 1.0, 1.0])
    assert np.allclose(mock_stdev.call_args_list[0][0][2], [0.0, 0.2, 0.4, 0.125])
    assert np.allclose(
        mock_stdev.call_args_list[0][0][3], np.diagflat([1.0, 1.0, 1.0, 1.0])
    )
    assert mock_stdev.call_args_list[0][1] == {}

    # post-fit prediction, single-channel model
    model = pyhf.Workspace(example_spec).model()
    fit_results = FitResults(
        np.asarray([1.1, 1.01]),
        np.asarray([0.1, 0.03]),
        ["Signal strength", "staterror_Signal-Region[0]"],
        np.asarray([[1.0, 0.2], [0.2, 1.0]]),
        0.0,
    )
    model_pred = model_utils.prediction(model, fit_results=fit_results)
    assert model_pred.model == model
    assert np.allclose(model_pred.model_yields, [[[57.54980000]]])  # new par value
    assert model_pred.total_stdev_model_bins == [[[0.3], [0.3]]]  # from mock
    assert model_pred.total_stdev_model_channels == [[0.3, 0.3]]  # from mock
    assert model_pred.label == "post-fit"
    assert "parameter names in fit results and model do not match" not in [
        rec.message for rec in caplog.records
    ]

    assert mock_asimov.call_count == 1  # no new call
    assert mock_unc.call_count == 1  # no new call

    # call to stdev calculation with fit_results propagated
    assert mock_stdev.call_count == 2
    assert mock_stdev.call_args_list[1][0][0] == model
    assert np.allclose(mock_stdev.call_args_list[1][0][1], [1.1, 1.01])
    assert np.allclose(mock_stdev.call_args_list[1][0][2], [0.1, 0.03])
    assert np.allclose(
        mock_stdev.call_args_list[1][0][3], np.asarray([[1.0, 0.2], [0.2, 1.0]])
    )
    assert mock_stdev.call_args_list[1][1] == {}

    caplog.clear()

    # custom prediction label, mismatch in parameter names
    fit_results = FitResults(
        np.asarray([1.1, 1.01]),
        np.asarray([0.1, 0.03]),
        ["a", "b"],
        np.asarray([[1.0, 0.2], [0.2, 1.0]]),
        0.0,
    )
    model_pred = model_utils.prediction(model, fit_results=fit_results, label="abc")
    assert "parameter names in fit results and model do not match" in [
        rec.message for rec in caplog.records
    ]
    assert model_pred.label == "abc"
    caplog.clear()


def test_unconstrained_parameter_count(example_spec, example_spec_shapefactor):
    model = pyhf.Workspace(example_spec).model()
    assert model_utils.unconstrained_parameter_count(model) == 1

    model = pyhf.Workspace(example_spec_shapefactor).model()
    assert model_utils.unconstrained_parameter_count(model) == 3

    # fixed parameters are skipped in counting
    example_spec_shapefactor["measurements"][0]["config"]["parameters"].append(
        {"name": "Signal strength", "fixed": True}
    )
    model = pyhf.Workspace(example_spec_shapefactor).model()
    assert model_utils.unconstrained_parameter_count(model) == 2


def test__parameter_index(caplog):
    caplog.set_level(logging.DEBUG)
    labels = ["a", "b", "c"]
    par_name = "b"
    assert model_utils._parameter_index(par_name, labels) == 1
    assert model_utils._parameter_index(par_name, tuple(labels)) == 1
    caplog.clear()

    assert model_utils._parameter_index("x", labels) is None
    assert "parameter x not found in model" in [rec.message for rec in caplog.records]
    caplog.clear()


def test__poi_index(example_spec, caplog):
    caplog.set_level(logging.DEBUG)
    model = pyhf.Workspace(example_spec).model()

    # POI given by name
    assert model_utils._poi_index(model, poi_name="staterror_Signal-Region[0]") == 1

    # parameter not found in model
    with pytest.raises(ValueError, match="parameter x not found in model"):
        model_utils._poi_index(model, poi_name="x")

    # POI specified in model config
    assert model_utils._poi_index(model) == 0
    caplog.clear()

    # no POI
    example_spec["measurements"][0]["config"]["poi"] = ""
    model = pyhf.Workspace(example_spec).model()
    assert model_utils._poi_index(model) is None
    assert "could not find POI for model" in [rec.message for rec in caplog.records]
    caplog.clear()


def test__strip_auxdata(example_spec):
    model = pyhf.Workspace(example_spec).model()
    data_with_aux = list(model.expected_data([1.0, 1.0], include_auxdata=True))
    data_without_aux = list(model.expected_data([1.0, 1.0], include_auxdata=False))

    assert model_utils._strip_auxdata(model, data_with_aux) == [51.8]
    assert model_utils._strip_auxdata(model, data_without_aux) == [51.8]


@mock.patch("cabinetry.model_utils._strip_auxdata", return_value=[25.0, 5.0, 8.0])
def test__data_per_channel(mock_aux, example_spec_multibin):
    model = pyhf.Workspace(example_spec_multibin).model()
    data = [25.0, 5.0, 8.0, 1.0, 1.0, 1.0]

    data_per_ch = model_utils._data_per_channel(model, [25.0, 5.0, 8.0, 1.0, 1.0, 1.0])
    assert data_per_ch == [[25.0, 5.0], [8.0]]  # auxdata stripped and split by channel

    # auxdata and channel index call
    assert mock_aux.call_args_list == [((model, data), {})]


def test__filter_channels(example_spec_multibin, caplog):
    caplog.set_level(logging.DEBUG)
    model = pyhf.Workspace(example_spec_multibin).model()

    assert model_utils._filter_channels(model, None) == ["region_1", "region_2"]
    assert model_utils._filter_channels(model, "region_1") == ["region_1"]
    assert model_utils._filter_channels(model, ["region_1", "region_2"]) == [
        "region_1",
        "region_2",
    ]
    assert model_utils._filter_channels(model, ["region_1", "abc"]) == ["region_1"]
    caplog.clear()

    # no matching channels
    assert model_utils._filter_channels(model, "abc") == []
    assert (
        "channel(s) ['abc'] not found in model, available channel(s): "
        "['region_1', 'region_2']" in [rec.message for rec in caplog.records]
    )
    caplog.clear()


@mock.patch(
    "cabinetry.model_utils.prefit_uncertainties",
    side_effect=[[0.4, 0.5, 0.6], [0.4, 0.5], [0.4, 0.5, 0.6]],
)
@mock.patch(
    "cabinetry.model_utils.asimov_parameters",
    side_effect=[[4.0, 5.0, 6.0], [4.0, 5.0], [4.0, 5.0, 6.0]],
)
def test_match_fit_results(mock_pars, mock_uncs):
    mock_model = mock.MagicMock()
    fit_results = FitResults(
        np.asarray([1.0, 2.0, 3.0]),
        np.asarray([0.1, 0.2, 0.3]),
        ["par_a", "par_b", "par_c"],
        np.asarray([[1.0, 0.2, 0.5], [0.2, 1.0, 0.1], [0.5, 0.1, 1.0]]),
        5.0,
        0.1,
    )

    # remove par_a, flip par_b and par_c, add par_d
    mock_model.config.par_names = ["par_c", "par_d", "par_b"]
    matched_fit_res = model_utils.match_fit_results(mock_model, fit_results)
    assert mock_pars.call_args_list == [((mock_model,), {})]
    assert mock_uncs.call_args_list == [((mock_model,), {})]
    assert np.allclose(matched_fit_res.bestfit, [3.0, 5.0, 2.0])
    assert np.allclose(matched_fit_res.uncertainty, [0.3, 0.5, 0.2])
    assert matched_fit_res.labels == ["par_c", "par_d", "par_b"]
    assert np.allclose(
        matched_fit_res.corr_mat, [[1.0, 0.0, 0.1], [0.0, 1.0, 0.0], [0.1, 0.0, 1.0]]
    )
    assert matched_fit_res.best_twice_nll == 5.0
    assert matched_fit_res.goodness_of_fit == 0.1

    # all parameters are new
    mock_model.config.par_names = ["par_d", "par_e"]
    matched_fit_res = model_utils.match_fit_results(mock_model, fit_results)
    assert np.allclose(matched_fit_res.bestfit, [4.0, 5.0])
    assert np.allclose(matched_fit_res.uncertainty, [0.4, 0.5])
    assert matched_fit_res.labels == ["par_d", "par_e"]
    assert np.allclose(matched_fit_res.corr_mat, [[1.0, 0.0], [0.0, 1.0]])
    assert matched_fit_res.best_twice_nll == 5.0
    assert matched_fit_res.goodness_of_fit == 0.1

    # fit results already match model exactly
    mock_model.config.par_names = ["par_a", "par_b", "par_c"]
    matched_fit_res = model_utils.match_fit_results(mock_model, fit_results)
    assert np.allclose(matched_fit_res.bestfit, [1.0, 2.0, 3.0])
    assert np.allclose(matched_fit_res.uncertainty, [0.1, 0.2, 0.3])
    assert matched_fit_res.labels == ["par_a", "par_b", "par_c"]
    assert np.allclose(
        matched_fit_res.corr_mat, [[1.0, 0.2, 0.5], [0.2, 1.0, 0.1], [0.5, 0.1, 1.0]]
    )
    assert matched_fit_res.best_twice_nll == 5.0
    assert matched_fit_res.goodness_of_fit == 0.1


def test_modifier_map(example_spec):
    model = pyhf.Workspace(example_spec).model()

    modifier_map = model_utils._modifier_map(model)
    assert modifier_map == {
        ("Signal Region", "Signal", "staterror_Signal-Region"): ["staterror"],
        ("Signal Region", "Signal", "Signal strength"): ["normfactor"],
    }
    assert modifier_map[("a", "b", "c")] == []


def test__parameters_maximizing_constraint_term(
    example_spec, example_spec_no_aux, example_spec_modifiers
):
    # staterror, custom auxdata value
    model = pyhf.Workspace(example_spec).model()
    best_pars = model_utils._parameters_maximizing_constraint_term(model, [1.1])
    assert np.allclose(best_pars, [1.0, 1.1])

    # no auxdata
    model = pyhf.Workspace(example_spec_no_aux).model()
    best_pars = model_utils._parameters_maximizing_constraint_term(model, [])
    assert np.allclose(best_pars, [1.0])

    # all modifiers (including Poisson constraint for shapesys), custom auxdata values
    # order: histosys, lumi, normfactor, normsys, shapefactor (2 bins),
    # shapesys (2 bins), staterror (2 bins)
    model = pyhf.Workspace(example_spec_modifiers).model()
    best_pars = model_utils._parameters_maximizing_constraint_term(
        model, [0.1, 1.1, 0.2, 1531.25, 281.25, 0.9, 0.8]
    )
    assert np.allclose(best_pars, [0.1, 1.1, 1.0, 0.2, 1.0, 1.0, 1.25, 1.25, 0.9, 0.8])



import logging

import numpy as np

from cabinetry import fit


def test_FitResults():
    bestfit = np.asarray([1.0])
    uncertainty = np.asarray([0.1])
    labels = ["par_a"]
    corr_mat = np.asarray([[1.0]])
    best_twice_nll = 2.0
    fit_results = fit.FitResults(bestfit, uncertainty, labels, corr_mat, best_twice_nll)
    assert np.allclose(fit_results.bestfit, bestfit)
    assert np.allclose(fit_results.uncertainty, uncertainty)
    assert fit_results.labels == labels
    assert np.allclose(fit_results.corr_mat, corr_mat)
    assert fit_results.best_twice_nll == best_twice_nll
    assert fit_results.goodness_of_fit == -1
    assert fit_results.minos_uncertainty == {}


def test_RankingResults():
    bestfit = np.asarray([1.0])
    uncertainty = np.asarray([0.1])
    labels = ["par_a"]
    prefit_up = np.asarray([0.3])
    prefit_down = np.asarray([-0.3])
    postfit_up = np.asarray([0.2])
    postfit_down = np.asarray([-0.2])
    ranking_results = fit.RankingResults(
        bestfit, uncertainty, labels, prefit_up, prefit_down, postfit_up, postfit_down
    )
    assert np.allclose(ranking_results.bestfit, bestfit)
    assert np.allclose(ranking_results.uncertainty, uncertainty)
    assert ranking_results.labels == labels
    assert np.allclose(ranking_results.prefit_up, prefit_up)
    assert np.allclose(ranking_results.prefit_down, prefit_down)
    assert np.allclose(ranking_results.postfit_up, postfit_up)
    assert np.allclose(ranking_results.postfit_down, postfit_down)


def test_ScanResults():
    name = "par_a"
    bestfit = 1.2
    uncertainty = 0.3
    parameter_values = np.asarray([0.9, 1.2, 1.5])
    delta_nlls = np.asarray([1.0, 0.0, 1.0])
    scan_results = fit.ScanResults(
        name, bestfit, uncertainty, parameter_values, delta_nlls
    )
    assert scan_results.name == name
    assert scan_results.bestfit == bestfit
    assert scan_results.uncertainty == uncertainty
    assert np.allclose(scan_results.parameter_values, parameter_values)
    assert np.allclose(scan_results.delta_nlls, delta_nlls)


def test_LimitResults():
    observed_limit = 3.0
    expected_limit = np.asarray([1.0, 2.0, 3.0, 4.0, 5.0])
    observed_CLs = np.asarray([0.05])
    expected_CLs = np.asarray([0.01, 0.02, 0.05, 0.07, 0.10])
    poi_values = np.asarray([3.0])
    confidence_level = 0.90
    limit_results = fit.LimitResults(
        observed_limit,
        expected_limit,
        observed_CLs,
        expected_CLs,
        poi_values,
        confidence_level,
    )
    assert limit_results.observed_limit == observed_limit
    assert np.allclose(limit_results.expected_limit, expected_limit)
    assert np.allclose(limit_results.observed_CLs, observed_CLs)
    assert np.allclose(limit_results.expected_CLs, expected_CLs)
    assert np.allclose(limit_results.poi_values, poi_values)
    assert np.allclose(limit_results.confidence_level, confidence_level)


def test_SignificanceResults():
    obs_p_val = 0.02
    obs_significance = 2
    exp_p_val = 0.16
    exp_significance = 1
    significance_results = fit.SignificanceResults(
        obs_p_val, obs_significance, exp_p_val, exp_significance
    )
    assert significance_results.observed_p_value == obs_p_val
    assert significance_results.observed_significance == obs_significance
    assert significance_results.expected_p_value == exp_p_val
    assert significance_results.expected_significance == exp_significance


def test_print_results(caplog):
    caplog.set_level(logging.DEBUG)

    bestfit = np.asarray([1.0, 2.0])
    uncertainty = np.asarray([0.1, 0.3])
    labels = ["param_A", "param_B"]
    fit_results = fit.FitResults(bestfit, uncertainty, labels, np.empty(0), 0.0)

    fit.print_results(fit_results)
    assert "param_A =  1.0000 +/- 0.1000" in [rec.message for rec in caplog.records]
    assert "param_B =  2.0000 +/- 0.3000" in [rec.message for rec in caplog.records]
    caplog.clear()



import logging

import cabinetry


def test___dir__():
    assert dir(cabinetry) == sorted(cabinetry.__all__)


def test_set_logging(caplog):
    log = logging.getLogger("cabinetry")

    # message not recorded by default
    log.debug("log message")
    assert len(caplog.records) == 0
    caplog.clear()

    # set custom logging, now message is recorded
    cabinetry.set_logging()
    log.debug("log message")
    assert "log message" in [rec.message for rec in caplog.records]



"""Reads histograms with uproot."""

import logging

import boost_histogram as bh
import uproot


log = logging.getLogger(__name__)


def with_uproot(histo_path: str) -> bh.Histogram:
    """Reads a histogram with uproot and returns it.

    Args:
        histo_path (str): path to histogram, use a colon to distinguish between path to
            file and path to histogram within file (example: ``file.root:h1``)

    Returns:
        bh.Histogram: histogram containing data
    """
    log.debug(f"reading input histogram {histo_path}")
    hist = uproot.open(histo_path).to_boost()
    return hist



"""Creates histograms used as input for a minimal example of cabinetry."""
import os

import boost_histogram as bh
import numpy as np
import uproot


def run(output_directory):
    bins = [200.0, 300.0, 400.0, 500.0, 600.0]
    hist = bh.Histogram(bh.axis.Variable(bins), storage=bh.storage.Weight())
    file_name = output_directory + "/histograms.root"
    with uproot.recreate(file_name) as f:
        yields = np.asarray([112.0, 112.0, 124.0, 66.0])
        stdev = np.asarray([10.58300524, 10.58300524, 11.13552873, 8.1240384])
        hist[...] = np.stack([yields, stdev**2], axis=-1)
        f["SR/Data/Nominal"] = hist

        yields = np.asarray([0.0, 1.58536913, 23.6164268, 24.54892223])
        stdev = np.asarray([0.0, 0.19931166, 0.77410713, 0.78830503])
        hist[...] = np.stack([yields, stdev**2], axis=-1)
        f["SR/Signal/Nominal"] = hist

        yields = np.asarray([112.73896936, 128.62169539, 88.10700838, 55.24607072])
        stdev = np.asarray([4.76136678, 5.10645036, 4.21104367, 3.34933335])
        hist[...] = np.stack([yields, stdev**2], axis=-1)
        f["SR/Background/Nominal"] = hist

        yields = np.asarray([53.85246451, 85.0382603, 90.75880537, 78.14459379])
        stdev = np.asarray([3.29419354, 4.14450981, 4.27730978, 3.96701472])
        hist[...] = np.stack([yields, stdev**2], axis=-1)
        f["SR/Background/Modeling_Up"] = hist

        yields = np.asarray([98.13569365, 154.1222757, 135.20449815, 103.14744392])
        stdev = np.asarray([4.17025569, 6.14088444, 6.47920695, 6.2581315])
        hist[...] = np.stack([yields, stdev**2], axis=-1)
        f["SR/Background/WeightBasedModeling_Up"] = hist

        yields = np.asarray([78.91727855, 90.03518677, 61.67490587, 38.6722495])
        stdev = np.asarray([3.33295675, 3.57451525, 2.94773057, 2.34453334])
        hist[...] = np.stack([yields, stdev**2], axis=-1)
        f["SR/Background/WeightBasedModeling_Down"] = hist


if __name__ == "__main__":
    output_directory = "inputs/"
    if not os.path.exists(output_directory):
        os.mkdir(output_directory)
    run(output_directory)



"""Provides utilities for pyhf models."""

from collections import defaultdict
import json
import logging
from typing import Any, DefaultDict, Dict, List, NamedTuple, Optional, Tuple, Union

import numpy as np
import pyhf

from cabinetry.fit.results_containers import FitResults


log = logging.getLogger(__name__)


# cache holding results from yield uncertainty calculations
_YIELD_STDEV_CACHE: Dict[Any, Tuple[List[List[List[float]]], List[List[float]]]] = {}


class ModelPrediction(NamedTuple):
    """Model prediction with yields and total uncertainties per bin and channel.

    Args:
        model (pyhf.pdf.Model): model to which prediction corresponds to
        model_yields (List[List[List[float]]]): yields per channel, sample and bin,
            indices: channel, sample, bin
        total_stdev_model_bins (List[List[List[float]]]): total yield uncertainty per
            channel, sample and bin, indices: channel, sample, bin (last sample: sum
            over samples)
        total_stdev_model_channels (List[List[float]]): total yield uncertainty per
            channel and sample, indices: channel, sample (last sample: sum over samples)
        label (str): label for the prediction, e.g. "pre-fit" or "post-fit"
    """

    model: pyhf.pdf.Model
    model_yields: List[List[List[float]]]
    total_stdev_model_bins: List[List[List[float]]]
    total_stdev_model_channels: List[List[float]]
    label: str


def model_and_data(
    spec: Dict[str, Any], *, asimov: bool = False, include_auxdata: bool = True
) -> Tuple[pyhf.pdf.Model, List[float]]:
    """Returns model and data for a ``pyhf`` workspace specification.

    Args:
        spec (Dict[str, Any]): a ``pyhf`` workspace specification
        asimov (bool, optional): whether to return the Asimov dataset, defaults to False
        include_auxdata (bool, optional): whether to also return auxdata, defaults to
            True

    Returns:
        Tuple[pyhf.pdf.Model, List[float]]:
            - a HistFactory-style model in ``pyhf`` format
            - the data (plus auxdata if requested) for the model
    """
    workspace = pyhf.Workspace(spec)
    model = workspace.model(
        modifier_settings={
            "normsys": {"interpcode": "code4"},
            "histosys": {"interpcode": "code4p"},
        }
    )  # use HistFactory InterpCode=4 (default in pyhf since v0.6.0)
    if not asimov:
        data = workspace.data(model, include_auxdata=include_auxdata)
    else:
        data = asimov_data(model, include_auxdata=include_auxdata)
    return model, data


def asimov_data(
    model: pyhf.Model,
    *,
    fit_results: Optional[FitResults] = None,
    poi_name: Optional[str] = None,
    poi_value: Optional[float] = None,
    include_auxdata: bool = True,
) -> List[float]:
    """Returns the Asimov dataset (optionally with auxdata) for a model.

    Initial parameter settings for normalization factors in the workspace are treated as
    the default settings for that parameter. Fitting the Asimov dataset will recover
    these initial settings as the maximum likelihood estimate for normalization factors.
    Initial settings for other modifiers are ignored. If the fit_results keyword
    argument is used, the Asimov dataset is built to recover the fit results given when
    fitted again.

    Args:
        model (pyhf.Model): the model from which to construct the dataset
        fit_results (Optional[FitResults], optional): parameter configuration to use
            when building the Asimov dataset (using the best-fit result), defaults to
            None (then a pre-fit Asimov dataset is built)
        poi_name (Optional[str], optional): name of parameter to set to a custom value
            via poi_value, defaults to None (use POI specified in workspace)
        poi_value (Optional[float], optional): custom value to set POI specified via
            poi_name to, defaults to None (no custom value set)
        include_auxdata (bool, optional): whether to also return auxdata, defaults to
            True

    Returns:
        List[float]: the Asimov dataset
    """
    if fit_results is None:
        # pre-fit Asimov by default
        parameters = asimov_parameters(model)
    else:
        # Asimov data for given fit result (copy to not modify original fit_results)
        parameters = fit_results.bestfit.copy()

    if poi_value is not None:
        # set POI to custom value, using parameter specified in model or given in kwarg
        poi_index = _poi_index(model, poi_name=poi_name)
        if poi_index is None:
            raise ValueError(
                "no POI specified in model, use the poi_name argument to set POI name"
            )
        parameters[poi_index] = poi_value

    asimov_data = pyhf.tensorlib.tolist(
        model.expected_data(parameters, include_auxdata=include_auxdata)
    )
    return asimov_data


def asimov_parameters(model: pyhf.pdf.Model) -> np.ndarray:
    """Returns a list of Asimov parameter values for a model.

    For normfactors and shapefactors, initial parameter settings (specified in the
    workspace) are treated as nominal settings. This ignores custom auxiliary data set
    in the measurement configuration in the workspace.

    Args:
        model (pyhf.pdf.Model): model for which to extract the parameters

    Returns:
        np.ndarray: the Asimov parameters, in the same order as
        ``model.config.suggested_init()``
    """
    # create a list of Asimov parameters (constrained parameters at best-fit value from
    # the aux measurement, unconstrained parameters at init specified in the workspace)
    asimov_parameters = []
    for parameter in model.config.par_order:
        if not model.config.param_set(parameter).constrained:
            # unconstrained parameter: use suggested inits (for normfactor/shapefactor)
            inits = model.config.param_set(parameter).suggested_init
        elif dict(model.config.modifiers)[parameter] in ["histosys", "normsys"]:
            # histosys/normsys: Gaussian constraint, nominal value 0
            inits = [0.0] * model.config.param_set(parameter).n_parameters
        else:
            # remaining modifiers are staterror/lumi with Gaussian constraint, and
            # shapesys with Poisson constraint, all have nominal value of 1
            inits = [1.0] * model.config.param_set(parameter).n_parameters

        asimov_parameters += inits

    return np.asarray(asimov_parameters)


def prefit_uncertainties(model: pyhf.pdf.Model) -> np.ndarray:
    """Returns a list of pre-fit parameter uncertainties for a model.

    For unconstrained parameters the uncertainty is set to 0. It is also set to 0 for
    fixed parameters (similarly to how the post-fit uncertainties are defined to be 0).

    Args:
        model (pyhf.pdf.Model): model for which to extract the parameters

    Returns:
        np.ndarray: pre-fit uncertainties for the parameters, in the same order as
        ``model.config.suggested_init()``
    """
    pre_fit_unc = []  # pre-fit uncertainties for parameters
    for parameter in model.config.par_order:
        if model.config.param_set(parameter).constrained:
            # pre-fit uncertainty for constrained parameters (if fixed, set to 0.0)
            widths = [
                width if not fixed else 0.0
                for width, fixed in zip(
                    model.config.param_set(parameter).width(),
                    model.config.param_set(parameter).suggested_fixed,
                )
            ]
            pre_fit_unc += widths
        else:
            # unconstrained: normfactor or shapefactor, uncertainty is 0
            pre_fit_unc += [0.0] * model.config.param_set(parameter).n_parameters
    return np.asarray(pre_fit_unc)


def _hashable_model_key(
    model: pyhf.pdf.Model,
) -> Tuple[str, Tuple[Tuple[str, str], ...]]:
    """Compute a hashable representation of the values that uniquely identify a model.

    The ``pyhf.pdf.Model`` type is already hashable, but it uses the ``__hash__``
    inherited from ``object``, so a copy of a model has a distinct hash. The key
    returned by this function instead will hash to the same value for copies, but differ
    when the model represents a different likelihood.

    Note: The key returned here considers only the spec and interpolation codes. All
    other model configuration options leave it unchanged (e.g. ``poi_name``, overriding
    parameter bounds, etc.).

    Args:
        model (pyhf.model.Model): model to generate a key for

    Returns:
        Tuple[str, Tuple[Tuple[str, str], ...]]: a key that identifies the model
        by its spec and interpcodes
    """
    interpcodes = []
    for mod_type in sorted(model.config.modifier_settings.keys()):
        code = model.config.modifier_settings[mod_type]["interpcode"]
        interpcodes.append((mod_type, code))
    # sort since different orderings result in equivalent models,
    # but distinct strings
    spec_str = json.dumps(model.spec, sort_keys=True)
    return (spec_str, tuple(interpcodes))


def yield_stdev(
    model: pyhf.pdf.Model,
    parameters: np.ndarray,
    uncertainty: np.ndarray,
    corr_mat: np.ndarray,
) -> Tuple[List[List[List[float]]], List[List[float]]]:
    """Calculates symmetrized model yield standard deviation per channel / sample / bin.

    Returns both the uncertainties per bin (in a list of channels and samples), and the
    uncertainty of the total yield per channel (again, for a list of channels and
    samples). To calculate the uncertainties for the total yield per channel, the
    function internally treats the sum of yields per channel like another channel with
    one bin. Similarly, the sum over samples is treated as another sample. The results
    of this function are cached to speed up subsequent calls with the same arguments.

    Args:
        model (pyhf.pdf.Model): the model for which to calculate the standard deviations
            for all bins
        parameters (np.ndarray): central values of model parameters
        uncertainty (np.ndarray): uncertainty of model parameters
        corr_mat (np.ndarray): correlation matrix

    Returns:
        Tuple[List[List[List[float]]], List[List[float]]]:
            - list of channels, each channel is a list of samples, and each sample a
              list of standard deviations per bin (the last sample corresponds to a sum
              over all samples)
            - list of standard deviations per channel, each channel is a list containing
              the standard deviations per sample (the last sample corresponds to a sum
              over all samples)
    """
    # check whether results are already stored in cache
    cached_results = _YIELD_STDEV_CACHE.get(
        (
            _hashable_model_key(model),
            tuple(parameters),
            tuple(uncertainty),
            corr_mat.data.tobytes(),
        ),
        None,
    )
    if cached_results is not None:
        # return results from cache
        return cached_results

    # the lists up_variations and down_variations will contain the model distributions
    # with all parameters varied individually within uncertainties
    # indices: variation, channel, sample, bin
    # following the channels contained in the model, there are additional entries with
    # yields summed per channel (internally treated like additional channels) to get the
    # per-channel uncertainties
    # in the same way, the total model prediction is following the list of individual
    # samples (and then treated like an additional sample)
    up_variations = []
    down_variations = []

    # calculate the model distribution for every parameter varied up and down
    # within the respective uncertainties
    for i_par in range(model.config.npars):
        # central parameter values, but one parameter varied within uncertainties
        up_pars = parameters.copy().astype(float)  # ensure float for correct addition
        up_pars[i_par] += uncertainty[i_par]
        down_pars = parameters.copy().astype(float)
        down_pars[i_par] -= uncertainty[i_par]

        # model distribution per sample with this parameter varied up
        up_comb = pyhf.tensorlib.to_numpy(
            model.main_model.expected_data(up_pars, return_by_sample=True)
        )
        # attach another entry with the total model prediction (sum over all samples)
        # indices: sample, bin
        up_comb = np.vstack((up_comb, np.sum(up_comb, axis=0)))
        # turn into list of channels (keep all samples, select correct bins per channel)
        # indices: channel, sample, bin
        up_yields_per_channel = [
            up_comb[:, model.config.channel_slices[ch]] for ch in model.config.channels
        ]
        # calculate list of yields summed per channel
        up_yields_channel_sum = np.stack(
            [
                np.sum(chan_yields, axis=-1, keepdims=True)
                for chan_yields in up_yields_per_channel
            ]
        )
        # reshape to drop bin axis, transpose to turn channel axis into new bin axis
        # (channel, sample, bin) -> (sample, bin) where "bin" becomes channel sums
        up_yields_channel_sum = up_yields_channel_sum.reshape(
            up_yields_channel_sum.shape[:-1]
        ).T

        # concatenate per-channel sums to up_comb (along bin axis)
        up_yields = np.concatenate((up_comb, up_yields_channel_sum), axis=1)
        # indices: variation, sample, bin
        up_variations.append(up_yields.tolist())

        # model distribution per sample with this parameter varied down
        down_comb = pyhf.tensorlib.to_numpy(
            model.main_model.expected_data(down_pars, return_by_sample=True)
        )
        # add total prediction (sum over samples)
        down_comb = np.vstack((down_comb, np.sum(down_comb, axis=0)))
        # turn into list of channels
        down_yields_per_channel = [
            down_comb[:, model.config.channel_slices[ch]]
            for ch in model.config.channels
        ]

        # calculate list of yields summed per channel
        down_yields_channel_sum = np.stack(
            [
                np.sum(chan_yields, axis=-1, keepdims=True)
                for chan_yields in down_yields_per_channel
            ]
        )
        down_yields_channel_sum = down_yields_channel_sum.reshape(
            down_yields_channel_sum.shape[:-1]
        ).T

        down_yields = np.concatenate((down_comb, down_yields_channel_sum), axis=1)
        down_variations.append(down_yields)

    # convert to numpy arrays for further processing
    up_variations_np = np.asarray(up_variations)
    down_variations_np = np.asarray(down_variations)

    # calculate symmetric uncertainties for all components
    # indices: variation, channel (last entries sums), sample (last entry sum), bin
    sym_uncs = (up_variations_np - down_variations_np) / 2

    # calculate total variance, indexed by sample, bin (per-channel numbers act like
    # additional channels with one bin each)
    if np.count_nonzero(corr_mat - np.diagflat(np.ones_like(parameters))) == 0:
        # no off-diagonal contributions from correlation matrix (e.g. pre-fit)
        total_variance = np.sum(np.power(sym_uncs, 2), axis=0)
    else:
        # full calculation including off-diagonal contributions
        # with v as vector of variations (each element contains yields under variation)
        # and M as correlation matrix, calculate variance as follows:
        # variance = sum_i sum_j v[i] * M[i, j] * v[j]
        # where the product between elements of v again is elementwise (multiplying bin
        # yields), and the final variance shape is the same as element of v (yield
        # uncertainties per sample, bin)

        # possible optimizations that could be considered here:
        #   - skipping staterror-staterror terms for per-bin calculation (orthogonal)
        #   - taking advantage of correlation matrix symmetry
        #   - (optional) skipping combinations with correlations below threshold

        # calculate M[i, j] * v[j] first
        # indices: pars (i), pars (j), sample, bin
        m_times_v = corr_mat[..., np.newaxis, np.newaxis] * sym_uncs[np.newaxis, ...]
        # now multiply by v[i] as well, indices: pars(i), pars(j), sample, bin
        v_times_m_times_v = sym_uncs[:, np.newaxis, ...] * m_times_v
        # finally perform sums over i and j, remaining indices: sample, bin
        total_variance = np.sum(np.sum(v_times_m_times_v, axis=1), axis=0)

    # get number of bins from model config
    n_bins = sum(model.config.channel_nbins.values())

    # convert to standard deviations per bin and per channel
    # add back outer channel dimension for per-bin uncertainty
    # indices: (channel, sample, bin)
    total_stdev_per_bin = [
        np.sqrt(total_variance[:, :n_bins][:, model.config.channel_slices[ch]]).tolist()
        for ch in model.config.channels
    ]
    # per-channel: transpose to flip remaining dimensions (channel sums acted as
    # individual bins before)
    # indices: (channel, sample)
    total_stdev_per_channel = np.sqrt(total_variance[:, n_bins:].T).tolist()

    # log total stdev per bin / channel (-1 index for sample sum)
    n_channels = len(model.config.channels)
    total_stdev_bin = [total_stdev_per_bin[i_chan][-1] for i_chan in range(n_channels)]
    log.debug(f"total stdev is {total_stdev_bin}")
    total_stdev_chan = [
        total_stdev_per_channel[i_chan][-1] for i_chan in range(n_channels)
    ]
    log.debug(f"total stdev per channel is {total_stdev_chan}")

    # save to cache
    _YIELD_STDEV_CACHE.update(
        {
            (
                _hashable_model_key(model),
                tuple(parameters),
                tuple(uncertainty),
                corr_mat.data.tobytes(),
            ): (total_stdev_per_bin, total_stdev_per_channel)
        }
    )

    return total_stdev_per_bin, total_stdev_per_channel


def prediction(
    model: pyhf.pdf.Model,
    *,
    fit_results: Optional[FitResults] = None,
    label: Optional[str] = None,
) -> ModelPrediction:
    """Returns model prediction, including model yields and uncertainties.

    If the optional fit result is not provided, the pre-fit Asimov yields and
    uncertainties are calculated. If the fit result is provided, the best-fit parameter
    values, uncertainties, and the parameter correlations are used to obtain the post-
    fit model and its associated uncertainties.

    Args:
        model (pyhf.pdf.Model): model to evaluate yield prediction for
        fit_results (Optional[FitResults], optional): parameter configuration to use,
            includes best-fit settings and uncertainties, as well as correlation matrix,
            defaults to None (then the pre-fit configuration is used)
        label (Optional[str], optional): label to include in model prediction, defaults
            to None (then will use "pre-fit" if fit results are not included, and "post-
            fit" otherwise)

    Returns:
        ModelPrediction: model, yields and uncertainties per channel, sample, bin
    """
    if fit_results is not None:
        if fit_results.labels != model.config.par_names:
            log.warning("parameter names in fit results and model do not match")
        # fit results specified, so they are used
        param_values = fit_results.bestfit
        param_uncertainty = fit_results.uncertainty
        corr_mat = fit_results.corr_mat
        label = "post-fit" if label is None else label

    else:
        # no fit results specified, generate pre-fit parameter values, uncertainties,
        # and diagonal correlation matrix
        param_values = asimov_parameters(model)
        param_uncertainty = prefit_uncertainties(model)
        corr_mat = np.zeros(shape=(len(param_values), len(param_values)))
        np.fill_diagonal(corr_mat, 1.0)
        label = "pre-fit" if label is None else label

    yields_combined = pyhf.tensorlib.to_numpy(
        model.main_model.expected_data(param_values, return_by_sample=True)
    )  # all channels concatenated

    # slice the yields into list of lists (of lists) where first index is channel,
    # second index is sample (and third index is bin)
    model_yields = [
        yields_combined[:, model.config.channel_slices[ch]].tolist()
        for ch in model.config.channels
    ]

    # calculate the total standard deviation of the model prediction
    # indices: (channel, sample, bin) for per-bin uncertainties,
    # (channel, sample) for per-channel
    total_stdev_model_bins, total_stdev_model_channels = yield_stdev(
        model, param_values, param_uncertainty, corr_mat
    )

    return ModelPrediction(
        model, model_yields, total_stdev_model_bins, total_stdev_model_channels, label
    )


def unconstrained_parameter_count(model: pyhf.pdf.Model) -> int:
    """Returns the number of unconstrained parameters in a model.

    The number is the sum of all independent parameters in a fit. A shapefactor that
    affects multiple bins enters the count once for each independent bin. Parameters
    that are set to constant are not included in the count.

    Args:
        model (pyhf.pdf.Model): model to count parameters for

    Returns:
        int: number of unconstrained parameters
    """
    n_pars = 0
    for parname in model.config.par_order:
        if not model.config.param_set(parname).constrained:
            # only consider non-constant parameters
            n_pars += model.config.param_set(parname).suggested_fixed.count(False)
    return n_pars


def _parameter_index(
    par_name: str, labels: Union[List[str], Tuple[str, ...]]
) -> Optional[int]:
    """Returns the position of a parameter with a given name in the list of parameters.

    Useful together with ``pyhf.pdf._ModelConfig.par_names`` to find the position of a
    parameter when the name is known. If the parameter is not found, logs an error and
    returns a default value of None.

    Args:
        par_name (str): name of parameter to find in list
        labels (Union[List[str], Tuple[str, ...]]): list or tuple with all parameter
            names in the model

    Returns:
        Optional[int]: index of parameter, or None if parameter was not found
    """
    par_index = next((i for i, label in enumerate(labels) if label == par_name), None)
    if par_index is None:
        log.error(f"parameter {par_name} not found in model")
    return par_index


def _poi_index(
    model: pyhf.pdf.Model, *, poi_name: Optional[str] = None
) -> Optional[int]:
    """Returns the index of the POI specified in the argument or the model default.

    If a string is given as argument, this takes priority. Otherwise the POI from the
    model is used. If no POI is specified there either, logs an error and returns a
    default value of None.

    Args:
        model (pyhf.pdf.Model): model for which to find the POI index
        poi_name (Optional[str], optional): name of the POI, defaults to None

    Raises:
        ValueError: if the specified POI name cannot be found in the model

    Returns:
        Optional[int]: POI index, or None if no POI could be found
    """
    if poi_name is not None:
        # use POI given by kwarg if specified
        poi_index = _parameter_index(poi_name, model.config.par_names)
        if poi_index is None:
            raise ValueError(f"parameter {poi_name} not found in model")
    elif model.config.poi_index is not None:
        # use POI specified in model
        poi_index = model.config.poi_index
    else:
        log.error("could not find POI for model")
        poi_index = None

    return poi_index


def _strip_auxdata(model: pyhf.pdf.Model, data: List[float]) -> List[float]:
    """Always returns observed yields, no matter whether data includes auxdata.

    Args:
        model (pyhf.pdf.Model): model to which data corresponds to
        data (List[float]): data, either including auxdata which is then stripped off or
            only observed yields

    Returns:
        List[float]: observed data yields
    """
    n_bins_total = sum(model.config.channel_nbins.values())
    if len(data) != n_bins_total:
        # strip auxdata, only observed yields are needed
        data = data[:n_bins_total]
    return data


def _data_per_channel(model: pyhf.pdf.Model, data: List[float]) -> List[List[float]]:
    """Returns data split per channel, and strips off auxiliary data if included.

    Args:
        model (pyhf.pdf.Model): model to which data corresponds to
        data (List[float]): data (not split by channel), can either include auxdata
            which is then stripped off, or only observed yields

    Returns:
        List[List[float]]: data per channel and per bin
    """
    # strip off auxiliary data
    data_combined = _strip_auxdata(model, data)

    # data is indexed by channel (and bin)
    data_yields = [
        data_combined[model.config.channel_slices[ch]] for ch in model.config.channels
    ]
    return data_yields


def _filter_channels(
    model: pyhf.pdf.Model, channels: Optional[Union[str, List[str]]]
) -> List[str]:
    """Returns a list of channels in a model after applying filtering.

    Args:
        model (pyhf.pdf.Model): model from which to extract channels
        channels (Optional[Union[str, List[str]]]): name of channel or list of channels
            to filter, only including those channels provided via this argument in the
            return of the function

    Returns:
        List[str]: list of channels after filtering
    """
    # channels included in model
    filtered_channels = model.config.channels
    # if one or more custom channels are provided, only include those
    if channels is not None:
        if isinstance(channels, str):
            channels = [channels]  # ensure list
        # only include if channel exists in model
        filtered_channels = [ch for ch in channels if ch in model.config.channels]

    if filtered_channels == []:
        log.warning(
            f"channel(s) {channels} not found in model, available channel(s): "
            f"{model.config.channels}"
        )

    return filtered_channels


def match_fit_results(model: pyhf.pdf.Model, fit_results: FitResults) -> FitResults:
    """Matches results from a fit to a model by adding or removing parameters as needed.

    If the fit results contain parameters missing in the model, these parameters are not
    included in the returned fit results. If the fit results do not include parameters
    used in the model, they are added to the fit results. The best-fit value for such
    parameters are the Asimov values as returned by ``asimov_parameters`` (initial
    parameter settings for unconstrained parameters), and the associated uncertainties
    as given by ``prefit_uncertainties`` (zero uncertainty for unconstrained or fixed
    parameters). These parameters furthermore are assumed to have no correlation with
    any other parameters. If required, parameters are re-ordered to match the target
    model.

    Args:
        model (pyhf.pdf.Model): model to match fit results to
        fit_results (FitResults): fit results to be updated in order to match model

    Returns:
        FitResults: fit results matching the model
    """
    # start with the assumption that the provided fit results contain no relevant
    # information for the target model at all, and initialize everything accordingly
    # then inject information contained in provided fit results and return the new
    # fit results matching the target model at the end

    bestfit = asimov_parameters(model)  # Asimov parameter values for target model
    uncertainty = prefit_uncertainties(model)  # pre-fit uncertainties for target model
    labels = model.config.par_names  # labels for target model

    # indices of parameters in current fit results, or None if they are missing
    indices_for_corr: List[Optional[int]] = [None] * len(labels)

    # loop over all required parameters
    for target_idx, target_label in enumerate(labels):
        # if parameters are missing in fit results, no further action is needed here
        # as all relevant objects have been initialized with that assumption
        if target_label in fit_results.labels:
            # fit results contain parameter, find its position
            idx_in_fit_results = fit_results.labels.index(target_label)
            # override pre-fit values by actual fit results for parameters
            bestfit[target_idx] = fit_results.bestfit[idx_in_fit_results]
            uncertainty[target_idx] = fit_results.uncertainty[idx_in_fit_results]
            # update indices for correlation matrix reconstruction
            indices_for_corr[target_idx] = idx_in_fit_results

    # re-build correlation matrix: start with diagonal matrix (assuming fit results do
    # not contain relevant info), and then insert values provided in fit results
    corr_mat = np.diagflat(np.ones_like(labels, dtype=float))
    for i_target, i_orig in enumerate(indices_for_corr):
        for j_target, j_orig in enumerate(indices_for_corr):
            # i_target and j_target are positions in matched correlation matrix
            # i_orig and j_orig are positions in old matrix, None if missing from there
            if i_orig is not None and j_orig is not None:
                # if i_orig or j_orig are None, one of the parameters are not part of
                # original fit result, and no update to correlation matrix is needed
                corr_mat[i_target][j_target] = fit_results.corr_mat[i_orig][j_orig]

    fit_results_matched = FitResults(
        np.asarray(bestfit),
        np.asarray(uncertainty),
        labels,
        corr_mat,
        fit_results.best_twice_nll,
        fit_results.goodness_of_fit,
    )
    return fit_results_matched


def _modifier_map(
    model: pyhf.pdf.Model,
) -> DefaultDict[Tuple[str, str, str], List[str]]:
    """Creates a map for modifier lists per (channel, sample, parameter).

    Args:
        model (pyhf.pdf.Model): model for which to create the map

    Returns:
        Dict[Tuple[str, str, str], List[str]]: map to extract modifier lists for each
        (channel, sample, parameter).
    """
    modifier_map = defaultdict(list)
    for channel in model.spec["channels"]:
        for sample in channel["samples"]:
            for modifier in sample["modifiers"]:
                modifier_map[
                    (channel["name"], sample["name"], modifier["name"])
                ].append(modifier["type"])
    return modifier_map


def _parameters_maximizing_constraint_term(
    model: pyhf.pdf.Model, aux_data: List[float]
) -> List[float]:
    """Returns parameters maximizing the constraint term for the given auxiliary data.

    Parameters without an associated constraint term are set to their initial value.

    Args:
        model (pyhf.pdf.Model): model to use for constraint term evaluation
        aux_data (List[float]): auxiliary data for which the constraint term should be
            maximized

    Returns:
        List[float]: parameters maximizing the model constraint term
    """
    best_pars = []  # parameters maximizing constraint term
    i_aux = 0  # current position in auxiliary data list
    i_poisson = 0  # current position in list of Poisson rescale factors

    try:
        # factors to rescale Poisson auxiliary data (index [0] for batch_size)
        poisson_factors = model.constraint_model.constraints_poisson.batched_factors[0]
    except AttributeError:
        poisson_factors = None  # no terms with Poisson constraint

    for parname in model.config.par_order:
        if parname in model.config.auxdata_order:
            # parameter has associated auxdata: determine parameter value from it
            # handle parameters controlling multiple bins
            n_params = model.config.param_set(parname).n_parameters
            if model.config.param_set(parname).pdf_type == "poisson":
                # auxiliary data for Poisson terms needs rescaling
                rescale_factors = [
                    float(rf)
                    for rf in poisson_factors[i_poisson : i_poisson + n_params]
                ]
                i_poisson += n_params
            else:
                rescale_factors = [1.0] * n_params  # no rescaling by default

            best_pars += list(
                np.asarray(aux_data[i_aux : i_aux + n_params]) / rescale_factors
            )
            i_aux += n_params

        else:
            # no associated auxdata (free-floating parameters): use init value
            par_inits = pyhf.tensorlib.tolist(
                model.config.param_set(parname).suggested_init
            )
            # ensure conversion to float https://github.com/scikit-hep/pyhf/issues/2065
            best_pars += [float(par_init) for par_init in par_inits]
    return best_pars



"""Applies optional post-processing to template histograms."""

import copy
import logging
import pathlib
from typing import Any, Dict, Literal, Optional

import numpy as np

from cabinetry import configuration
from cabinetry import histo
from cabinetry import route
from cabinetry import smooth


log = logging.getLogger(__name__)


def _fix_stat_unc(histogram: histo.Histogram, name: str) -> None:
    """Replaces NaN statistical uncertainties by zero for a histogram.

    Modifies the histogram handed over in the argument.

    Args:
        histogram (cabinetry.histo.Histogram): the histogram to fix
        name (str): histogram name for logging
    """
    nan_pos = np.where(np.isnan(histogram.stdev))[0]
    if len(nan_pos) > 0:
        log.debug(f"fixing ill-defined stat. unc. for {name}")
        histogram.stdev = np.nan_to_num(histogram.stdev, nan=0.0)


def _apply_353qh_twice(
    variation: histo.Histogram, nominal: histo.Histogram, name: str
) -> None:
    """Smooths systematic template histogram with the "353QH, twice" algorithm.

    The algorithm is applied to the ratio systematic variation / nominal. The nominal
    histogram stays the same, while the variation histogram is modified. Statistical
    uncertainties do not change (and the algorithm is not aware of statistical
    uncertainties). The total yield of the variation histogram stays unchanged.

    Args:
        variation (histo.Histogram): histogram of systematic variation
        nominal (histo.Histogram): associated nominal histogram
        name (str): histogram name for logging
    """
    log.debug(f"applying smoothing to {name}")
    # smooth relative effect of systematic (systematic/nominal)
    smooth_var = (
        smooth.smooth_353qh_twice(variation.yields / nominal.yields) * nominal.yields
    )
    # scale to match original sum of yields of variation
    variation.yields = smooth_var * sum(variation.yields) / sum(smooth_var)


def _smoothing_algorithm(
    region: Dict[str, Any], sample: Dict[str, Any], systematic: Dict[str, Any]
) -> Optional[str]:
    """Returns name of algorithm to use for smoothing, or None otherwise.

    Args:
        region (Dict[str, Any]): containing all region information
        sample (Dict[str, Any]): containing all sample information
        systematic (Dict[str, Any]): containing all systematic information

    Returns:
        Optional[str]: name of smoothing algorithm or None
    """
    smoothing = systematic.get("Smoothing", None)
    if smoothing is None:
        return None

    # check for region and sample restrictions specified for smoothing, apply to all
    # regions and samples by default if no further specification is made
    if not configuration._x_contains_y(region, smoothing, "Regions"):
        # regions are specified and current region does not match, do not smooth
        return None

    if not configuration._x_contains_y(sample, smoothing, "Samples"):
        # samples are specified and current sample does not match, do not smooth
        return None

    # smoothing algorithm needs to be applied
    smoothing_algorithm = smoothing["Algorithm"]
    return smoothing_algorithm


def apply_postprocessing(
    histogram: histo.Histogram,
    name: str,
    *,
    smoothing_algorithm: Optional[str] = None,
    nominal_histogram: Optional[histo.Histogram] = None,
) -> histo.Histogram:
    """Returns a new histogram with post-processing applied.

    The histogram handed to the function stays unchanged. A copy of the histogram
    receives post-processing and is then returned. The postprocessing consists of a fix
    for NaN statistical uncertainties and optional smoothing.

    Args:
        histogram (cabinetry.histo.Histogram): the histogram to postprocess
        name (str): histogram name for logging
        smoothing_algorithm (Optional[str]): name of smoothing algorithm to apply,
            defaults to None (no smoothing done)
        nominal_histogram (Optional[cabinetry.histo.Histogram]): nominal histogram
            (needed for smoothing), defaults to None

    Returns:
        cabinetry.histo.Histogram: the histogram with post-processing applied
    """
    # copy histogram to new object to leave it unchanged
    modified_histogram = copy.deepcopy(histogram)
    _fix_stat_unc(modified_histogram, name)
    if smoothing_algorithm is not None:
        if smoothing_algorithm == "353QH, twice":
            if nominal_histogram is None:
                raise ValueError("cannot apply smoothing, nominal histogram missing")
            _apply_353qh_twice(modified_histogram, nominal_histogram, name)
        else:
            log.warning(f"unknown smoothing algorithm {smoothing_algorithm}")
    return modified_histogram


def _postprocessor(histogram_folder: pathlib.Path) -> route.ProcessorFunc:
    """Returns the post-processing function to be applied to template histograms.

    Needed by ``cabinetry.route.apply_to_all_templates``. Could alternatively create a
    ``Postprocessor`` class that contains processors.

    Args:
        histogram_folder (pathlib.Path): folder containing histograms

    Returns:
        route.ProcessorFunc: function to apply to a template histogram
    """

    def process_template(
        region: Dict[str, Any],
        sample: Dict[str, Any],
        systematic: Dict[str, Any],
        template: Optional[Literal["Up", "Down"]],
    ) -> None:
        """Applies post-processing to a single histogram.

        Args:
            region (Dict[str, Any]): containing all region information
            sample (Dict[str, Any]): containing all sample information
            systematic (Dict[str, Any]): containing all systematic information
            template (Optional[Literal["Up", "Down"]]): template considered: "Up",
                "Down", or None for nominal
        """
        histogram = histo.Histogram.from_config(
            histogram_folder,
            region,
            sample,
            systematic,
            template=template,
            modified=False,
        )
        histogram_name = histo.name(region, sample, systematic, template=template)

        smoothing_algorithm = _smoothing_algorithm(region, sample, systematic)
        if smoothing_algorithm is None:
            nominal_histogram = None
        else:
            # to apply smoothing, the associated nominal histogram is needed (smoothing
            # can currently not be applied to nominal itself, but for systematics is
            # applied to the ratio variation / nominal)
            nominal_histogram = histo.Histogram.from_config(
                histogram_folder, region, sample, {}, modified=False
            )

        new_histogram = apply_postprocessing(
            histogram,
            histogram_name,
            smoothing_algorithm=smoothing_algorithm,
            nominal_histogram=nominal_histogram,
        )
        histogram.validate(histogram_name)
        new_histo_path = histogram_folder / (histogram_name + "_modified")
        new_histogram.save(new_histo_path)

    return process_template



"""High-level entry point for statistical inference."""

import logging
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

import iminuit
import numpy as np
import pyhf
import scipy.optimize
import scipy.stats

from cabinetry import model_utils
from cabinetry.fit.results_containers import (
    FitResults,
    LimitResults,
    RankingResults,
    ScanResults,
    SignificanceResults,
)


log = logging.getLogger(__name__)


def print_results(fit_results: FitResults) -> None:
    """Prints the best-fit parameter results and associated uncertainties.

    Parameters with zero uncertainty (those held fixed in the fit) are in addition
    denoted with "(constant)".

    Args:
        fit_results (FitResults): results of fit to be printed
    """
    max_label_length = max(len(label) for label in fit_results.labels)
    log.info("fit results (with symmetric uncertainties):")
    for i, label in enumerate(fit_results.labels):
        const_label = "  (constant)" if fit_results.uncertainty[i] == 0.0 else ""
        log.info(
            f"{label:<{max_label_length}} = {fit_results.bestfit[i]: .4f} +/- "
            f"{fit_results.uncertainty[i]:.4f}{const_label}"
        )


def _fit_model_pyhf(
    model: pyhf.pdf.Model,
    data: List[float],
    *,
    minos: Optional[Union[List[str], Tuple[str, ...]]] = None,
    init_pars: Optional[List[float]] = None,
    fix_pars: Optional[List[bool]] = None,
    par_bounds: Optional[List[Tuple[float, float]]] = None,
    strategy: Optional[Literal[0, 1, 2]] = None,
    maxiter: Optional[int] = None,
    tolerance: Optional[float] = None,
) -> FitResults:
    """Uses the ``pyhf.infer`` API to perform a maximum likelihood fit.

    Parameters set to be fixed in the model are held constant. The ``init_pars``
    argument allows to override the ``pyhf`` default initial parameter settings, the
    ``fix_pars`` argument overrides which parameters are held constant, ``par_bounds``
    sets parameter bounds.

    Args:
        model (pyhf.pdf.Model): the model to use in the fit
        data (List[float]): the data to fit the model to
        minos (Optional[Union[List[str], Tuple[str, ...]]], optional): runs the MINOS
            algorithm for all parameters specified, defaults to None (does not run
            MINOS)
        init_pars (Optional[List[float]], optional): list of initial parameter settings,
            defaults to None (use ``pyhf`` suggested inits)
        fix_pars (Optional[List[bool]], optional): list of booleans specifying which
            parameters are held constant, defaults to None (use ``pyhf`` suggestion)
        par_bounds (Optional[List[Tuple[float, float]]], optional): list of tuples with
            parameter bounds for fit, defaults to None (use ``pyhf`` suggested bounds)
        strategy (Optional[Literal[0, 1, 2]], optional): minimization strategy used by
            Minuit, can be 0/1/2, defaults to None (then uses ``pyhf`` default behavior
            of strategy 0 with user-provided gradients and 1 otherwise)
        maxiter (Optional[int], optional): allowed number of calls for minimization,
            defaults to None (use ``pyhf`` default of 100,000)
        tolerance (Optional[float]), optional): tolerance for convergence, for details
            see ``iminuit.Minuit.tol`` (uses EDM < 0.002*tolerance), defaults to
            None (use ``iminuit`` default of 0.1)

    Returns:
        FitResults: object storing relevant fit results
    """
    _, initial_optimizer = pyhf.get_backend()  # store initial optimizer settings
    pyhf.set_backend(pyhf.tensorlib, pyhf.optimize.minuit_optimizer(verbose=1))

    # strategy=None is currently not supported in pyhf
    # https://github.com/scikit-hep/pyhf/issues/1785
    strategy_kwarg = {"strategy": strategy} if strategy is not None else {}

    result, corr_mat, best_twice_nll, result_obj = pyhf.infer.mle.fit(
        data,
        model,
        init_pars=init_pars,
        fixed_params=fix_pars,
        par_bounds=par_bounds,
        return_uncertainties=True,
        return_correlations=True,
        return_fitted_val=True,
        return_result_obj=True,
        maxiter=maxiter,
        tolerance=tolerance,
        **strategy_kwarg,
    )
    log.info(f"Migrad status:\n{result_obj.minuit.fmin}")

    bestfit = pyhf.tensorlib.to_numpy(result[:, 0])
    # set errors for fixed parameters to 0 (see iminuit#762)
    uncertainty = np.where(
        result_obj.minuit.fixed, 0.0, pyhf.tensorlib.to_numpy(result[:, 1])
    )
    labels = model.config.par_names
    corr_mat = pyhf.tensorlib.to_numpy(corr_mat)
    best_twice_nll = float(best_twice_nll)  # convert 0-dim np.ndarray to float

    minos_results = (
        _run_minos(result_obj.minuit, minos, labels) if minos is not None else {}
    )

    fit_results = FitResults(
        bestfit,
        uncertainty,
        labels,
        corr_mat,
        best_twice_nll,
        minos_uncertainty=minos_results,
    )
    pyhf.set_backend(pyhf.tensorlib, initial_optimizer)  # restore optimizer settings
    return fit_results


def _fit_model_custom(
    model: pyhf.pdf.Model,
    data: List[float],
    *,
    minos: Optional[Union[List[str], Tuple[str, ...]]] = None,
    init_pars: Optional[List[float]] = None,
    fix_pars: Optional[List[bool]] = None,
    par_bounds: Optional[List[Tuple[float, float]]] = None,
    strategy: Optional[Literal[0, 1, 2]] = None,
    maxiter: Optional[int] = None,
    tolerance: Optional[float] = None,
) -> FitResults:
    """Uses ``iminuit`` directly to perform a maximum likelihood fit.

    Parameters set to be fixed in the model are held constant. The ``init_pars``
    argument allows to override the ``pyhf`` default initial parameter settings, the
    ``fix_pars`` argument overrides which parameters are held constant, ``par_bounds``
    sets parameter bounds.

    Args:
        model (pyhf.pdf.Model): the model to use in the fit
        data (List[float]): the data to fit the model to
        minos (Optional[Union[List[str], Tuple[str, ...]]], optional): runs the MINOS
            algorithm for all parameters specified, defaults to None (does not run
            MINOS)
        init_pars (Optional[List[float]], optional): list of initial parameter settings,
            defaults to None (use ``pyhf`` suggested inits)
        fix_pars (Optional[List[bool]], optional): list of booleans specifying which
            parameters are held constant, defaults to None (use ``pyhf`` suggestion)
        par_bounds (Optional[List[Tuple[float, float]]], optional): list of tuples with
            parameter bounds for fit, defaults to None (use ``pyhf`` suggested bounds)
        strategy (Optional[Literal[0, 1, 2]], optional): minimization strategy used by
            Minuit, can be 0/1/2, defaults to None (then uses ``pyhf`` default behavior
            of strategy 0 with user-provided gradients and 1 otherwise)
        maxiter (Optional[int], optional): allowed number of calls for minimization,
            defaults to None (use ``pyhf`` default of 100,000)
        tolerance (Optional[float]), optional): tolerance for convergence, for details
            see ``iminuit.Minuit.tol`` (uses EDM < 0.002*tolerance), defaults to
            None (use ``iminuit`` default of 0.1)

    Raises:
        ValueError: if minimization fails

    Returns:
        FitResults: object storing relevant fit results
    """
    _, initial_optimizer = pyhf.get_backend()  # store initial optimizer settings
    pyhf.set_backend(pyhf.tensorlib, pyhf.optimize.minuit_optimizer(verbose=1))

    # use parameter settings provided in function arguments if they exist, else defaults
    init_pars = init_pars or model.config.suggested_init()
    fix_pars = fix_pars or model.config.suggested_fixed()
    par_bounds = par_bounds or model.config.suggested_bounds()

    labels = model.config.par_names

    def twice_nll_func(pars: np.ndarray) -> Any:
        """The objective for minimization: twice the negative log-likelihood.

        The return value is float-like, but not always a float. The actual type depends
        on the active ``pyhf`` backend.

        Args:
            pars (np.ndarray): parameter values at which the NLL is evaluated

        Returns:
            Any: twice the negative log-likelihood
        """
        twice_nll = -2 * model.logpdf(pars, data)
        return twice_nll[0]

    m = iminuit.Minuit(twice_nll_func, init_pars, name=labels)
    m.fixed = fix_pars
    m.limits = par_bounds
    m.errordef = 1
    m.print_level = 1

    if strategy is not None:
        m.strategy = strategy
    else:
        # pick strategy like pyhf: 0 if backend provides autodiff gradients, otherwise 1
        m.strategy = 0 if pyhf.tensorlib.default_do_grad else 1

    maxiter = maxiter or 100_000
    m.tol = tolerance or 0.1  # goal: EDM < 0.002*tol*errordef

    m.migrad(ncall=maxiter)
    m.hesse()  # use default call limit (consistent with pyhf)

    log.info(f"MINUIT status:\n{m.fmin}")
    if not m.valid:
        raise ValueError("Minimization failed, minimum is invalid.")

    bestfit = np.asarray(m.values)
    # set errors for fixed parameters to 0 (see iminuit#762)
    uncertainty = np.where(m.fixed, 0.0, m.errors)
    corr_mat = m.covariance.correlation()  # iminuit.util.Matrix, subclass of np.ndarray
    best_twice_nll = m.fval

    minos_results = _run_minos(m, minos, labels) if minos is not None else {}

    fit_results = FitResults(
        bestfit,
        uncertainty,
        labels,
        corr_mat,
        best_twice_nll,
        minos_uncertainty=minos_results,
    )
    pyhf.set_backend(pyhf.tensorlib, initial_optimizer)  # restore optimizer settings
    return fit_results


def _fit_model(
    model: pyhf.pdf.Model,
    data: List[float],
    *,
    minos: Optional[Union[List[str], Tuple[str, ...]]] = None,
    init_pars: Optional[List[float]] = None,
    fix_pars: Optional[List[bool]] = None,
    par_bounds: Optional[List[Tuple[float, float]]] = None,
    strategy: Optional[Literal[0, 1, 2]] = None,
    maxiter: Optional[int] = None,
    tolerance: Optional[float] = None,
    custom_fit: bool = False,
) -> FitResults:
    """Interface for maximum likelihood fits through ``pyhf.infer`` API or ``iminuit``.

    Parameters set to be fixed in the model are held constant. The ``init_pars``
    argument allows to override the ``pyhf`` default initial parameter settings, the
    ``fix_pars`` argument overrides which parameters are held constant, ``par_bounds``
    sets parameter bounds.

    Args:
        model (pyhf.pdf.Model): the model to use in the fit
        data (List[float]): the data to fit the model to
        minos (Optional[Union[List[str], Tuple[str, ...]]], optional): runs the MINOS
            algorithm for all parameters specified, defaults to None (does not run
            MINOS)
        init_pars (Optional[List[float]], optional): list of initial parameter settings,
            defaults to None (use ``pyhf`` suggested inits)
        fix_pars (Optional[List[bool]], optional): list of booleans specifying which
            parameters are held constant, defaults to None (use ``pyhf`` suggestion)
        par_bounds (Optional[List[Tuple[float, float]]], optional): list of tuples with
            parameter bounds for fit, defaults to None (use ``pyhf`` suggested bounds)
        strategy (Optional[Literal[0, 1, 2]], optional): minimization strategy used by
            Minuit, can be 0/1/2, defaults to None (then uses ``pyhf`` default behavior
            of strategy 0 with user-provided gradients and 1 otherwise)
        maxiter (Optional[int], optional): allowed number of calls for minimization,
            defaults to None (use ``pyhf`` default of 100,000)
        tolerance (Optional[float]), optional): tolerance for convergence, for details
            see ``iminuit.Minuit.tol`` (uses EDM < 0.002*tolerance), defaults to
            None (use ``iminuit`` default of 0.1)
        custom_fit (bool, optional): whether to use the ``pyhf.infer`` API or
            ``iminuit``, defaults to False (using ``pyhf.infer``)

    Returns:
        FitResults: object storing relevant fit results
    """
    if not custom_fit:
        # use pyhf infer API
        fit_results = _fit_model_pyhf(
            model,
            data,
            minos=minos,
            init_pars=init_pars,
            fix_pars=fix_pars,
            par_bounds=par_bounds,
            strategy=strategy,
            maxiter=maxiter,
            tolerance=tolerance,
        )
    else:
        # use iminuit directly
        fit_results = _fit_model_custom(
            model,
            data,
            minos=minos,
            init_pars=init_pars,
            fix_pars=fix_pars,
            par_bounds=par_bounds,
            strategy=strategy,
            maxiter=maxiter,
            tolerance=tolerance,
        )
    log.debug(f"-2 log(L) = {fit_results.best_twice_nll:.6f} at best-fit point")
    return fit_results


def _run_minos(
    minuit_obj: iminuit.Minuit,
    minos: Union[List[str], Tuple[str, ...]],
    labels: List[str],
) -> Dict[str, Tuple[float, float]]:
    """Determines parameter uncertainties for a list of parameters with MINOS.

    Args:
        minuit_obj (iminuit.Minuit): Minuit instance to use
        minos (Union[List[str], Tuple[str, ...]]): parameters for which MINOS is run
        labels (List[str]]): names of all parameters known to ``iminuit``, these names
            are used in output (may be the same as the names under which ``iminiuit``
            knows parameters)

    Returns:
        Dict[str, Tuple[float, float]]: uncertainties indexed by parameter name
    """
    for par_name in minos:
        if par_name not in minuit_obj.parameters:
            # parameter not found, skip calculation
            log.warning(f"parameter {par_name} not found in model")
            continue
        log.info(f"running MINOS for {par_name}")
        minuit_obj.minos(par_name)

    minos_results = {}

    log.info("MINOS results:")
    max_label_length = max(len(label) for label in labels)
    minos_unc = [minuit_obj.params[i].merror for i in range(minuit_obj.npar)]
    for i_par, unc in zip(range(len(labels)), minos_unc):
        # if MINOS has not been run, entries are None
        if unc is not None:
            log.info(
                f"{labels[i_par]:<{max_label_length}} = "
                f"{minuit_obj.values[i_par]: .4f} {unc[0]:+.4f} {unc[1]:+.4f}"
            )
            minos_results.update({labels[i_par]: (unc[0], unc[1])})

    return minos_results


def _goodness_of_fit(
    model: pyhf.pdf.Model, data: List[float], best_twice_nll: float
) -> float:
    """Calculates goodness-of-fit p-value with a saturated model.

    Returns NaN if the number of degrees of freedom in the chi2 test is zero (nominal
    fit should already be perfect) or negative (over-parameterized model). Does not
    require to run an additional fit to get the saturated model likelihood.

    Args:
        model (pyhf.pdf.Model): model used in the fit for which goodness-of-fit should
            be calculated
        data (List[float]): the observed data
        best_twice_nll (float): best-fit -2 log(likelihood) of fit for which goodness-
            of-fit should be calculated

    Returns:
        float: goodness-of-fit p-value
    """
    if model.config.nauxdata > 0:
        main_data, aux_data = model.fullpdf_tv.split(pyhf.tensorlib.astensor(data))

        # need to obtain the parameters that maximize the constraint term
        # (will not match suggested_init() when auxdata is custom)
        best_pars = model_utils._parameters_maximizing_constraint_term(
            model, pyhf.tensorlib.tolist(aux_data)
        )

        # constraint term: log Gaussian(aux_data|parameters) etc.
        constraint_ll = pyhf.tensorlib.to_numpy(
            model.constraint_logpdf(aux_data, pyhf.tensorlib.astensor(best_pars))
        )
    else:
        # no auxiliary data, so no constraint terms present
        main_data = pyhf.tensorlib.astensor(data)
        constraint_ll = 0.0
    # Poisson term: log Poisson(data|lambda=data), sum is over log likelihood of bins
    poisson_ll = pyhf.tensorlib.to_numpy(
        sum(pyhf.tensorlib.poisson_dist(main_data).log_prob(main_data))
    )
    saturated_nll = -(poisson_ll + constraint_ll)  # saturated likelihood

    log.info("calculating goodness-of-fit")
    delta_nll = best_twice_nll / 2 - saturated_nll
    log.debug(f"Delta NLL = {delta_nll:.6f}")

    # calculate difference in degrees of freedom between fits, given by the number
    # of bins minus the number of unconstrained parameters
    n_dof = sum(
        model.config.channel_nbins.values()
    ) - model_utils.unconstrained_parameter_count(model)
    log.debug(f"number of degrees of freedom: {n_dof}")

    if n_dof <= 0:
        log.warning(
            f"cannot calculate p-value: {n_dof} degrees of freedom and Delta NLL = "
            f"{delta_nll:.6f}"
        )
        return np.nan

    p_val = scipy.stats.chi2.sf(2 * delta_nll, n_dof)
    log.info(f"p-value for goodness-of-fit test: {p_val:.2%}")
    return p_val


def fit(
    model: pyhf.pdf.Model,
    data: List[float],
    *,
    minos: Optional[Union[str, List[str], Tuple[str, ...]]] = None,
    goodness_of_fit: bool = False,
    init_pars: Optional[List[float]] = None,
    fix_pars: Optional[List[bool]] = None,
    par_bounds: Optional[List[Tuple[float, float]]] = None,
    strategy: Optional[Literal[0, 1, 2]] = None,
    maxiter: Optional[int] = None,
    tolerance: Optional[float] = None,
    custom_fit: bool = False,
) -> FitResults:
    """Performs a  maximum likelihood fit, reports and returns the results.

    Depending on the ``custom_fit`` keyword argument, this uses either the
    ``pyhf.infer`` API or ``iminuit`` directly.

    Args:
        model (pyhf.pdf.Model): model to use in fit
        data (List[float]): data (including auxdata) the model is fit to
        minos (Optional[Union[str, List[str], Tuple[str, ...]]], optional): runs the
            MINOS algorithm for all parameters specified, defaults to None (does not run
            MINOS)
        goodness_of_fit (bool, optional): calculate goodness of fit with a saturated
            model (perfectly fits data with shapefactors in all bins), defaults to False
        init_pars (Optional[List[float]], optional): list of initial parameter settings,
            defaults to None (use ``pyhf`` suggested inits)
        fix_pars (Optional[List[bool]], optional): list of booleans specifying which
            parameters are held constant, defaults to None (use ``pyhf`` suggestion)
        par_bounds (Optional[List[Tuple[float, float]]], optional): list of tuples with
            parameter bounds for fit, defaults to None (use ``pyhf`` suggested bounds)
        strategy (Optional[Literal[0, 1, 2]], optional): minimization strategy used by
            Minuit, can be 0/1/2, defaults to None (then uses ``pyhf`` default behavior
            of strategy 0 with user-provided gradients and 1 otherwise)
        maxiter (Optional[int], optional): allowed number of calls for minimization,
            defaults to None (use ``pyhf`` default of 100,000)
        tolerance (Optional[float]), optional): tolerance for convergence, for details
            see ``iminuit.Minuit.tol`` (uses EDM < 0.002*tolerance), defaults to
            None (use ``iminuit`` default of 0.1)
        custom_fit (bool, optional): whether to use the ``pyhf.infer`` API or
            ``iminuit``, defaults to False (using ``pyhf.infer``)

    Returns:
        FitResults: object storing relevant fit results
    """
    log.info("performing maximum likelihood fit")

    # convert minos parameter to list if a single parameter is specified as string
    if isinstance(minos, str):
        minos = [minos]

    # perform fit
    fit_results = _fit_model(
        model,
        data,
        minos=minos,
        init_pars=init_pars,
        fix_pars=fix_pars,
        par_bounds=par_bounds,
        strategy=strategy,
        maxiter=maxiter,
        tolerance=tolerance,
        custom_fit=custom_fit,
    )

    print_results(fit_results)

    if goodness_of_fit:
        # calculate goodness-of-fit with saturated model
        p_val = _goodness_of_fit(model, data, fit_results.best_twice_nll)
        fit_results = fit_results._replace(goodness_of_fit=p_val)

    return fit_results


def ranking(
    model: pyhf.pdf.Model,
    data: List[float],
    *,
    fit_results: Optional[FitResults] = None,
    poi_name: Optional[str] = None,
    init_pars: Optional[List[float]] = None,
    fix_pars: Optional[List[bool]] = None,
    par_bounds: Optional[List[Tuple[float, float]]] = None,
    strategy: Optional[Literal[0, 1, 2]] = None,
    maxiter: Optional[int] = None,
    tolerance: Optional[float] = None,
    custom_fit: bool = False,
) -> RankingResults:
    """Calculates the impact of nuisance parameters on the parameter of interest (POI).

    The impact is given by the difference in the POI between the nominal fit, and a fit
    where the nuisance parameter is held constant at its nominal value plus/minus its
    associated uncertainty. The "pre-fit impact" is obtained by varying the nuisance
    parameters by their uncertainty given by their constraint term.

    Args:
        model (pyhf.pdf.Model): model to use in fits
        data (List[float]): data (including auxdata) the model is fit to
        fit_results (Optional[FitResults], optional): nominal fit results to use for
            ranking, if not specified will repeat nominal fit, defaults to None
        poi_name (Optional[str], optional): impact is calculated with respect to this
            parameter, defaults to None (use POI specified in workspace)
        init_pars (Optional[List[float]], optional): list of initial parameter settings,
            defaults to None (use ``pyhf`` suggested inits)
        fix_pars (Optional[List[bool]], optional): list of booleans specifying which
            parameters are held constant, defaults to None (use ``pyhf`` suggestion)
        par_bounds (Optional[List[Tuple[float, float]]], optional): list of tuples with
            parameter bounds for fit, defaults to None (use ``pyhf`` suggested bounds)
        strategy (Optional[Literal[0, 1, 2]], optional): minimization strategy used by
            Minuit, can be 0/1/2, defaults to None (then uses ``pyhf`` default behavior
            of strategy 0 with user-provided gradients and 1 otherwise)
        maxiter (Optional[int], optional): allowed number of calls for minimization,
            defaults to None (use ``pyhf`` default of 100,000)
        tolerance (Optional[float]), optional): tolerance for convergence, for details
            see ``iminuit.Minuit.tol`` (uses EDM < 0.002*tolerance), defaults to
            None (use ``iminuit`` default of 0.1)
        custom_fit (bool, optional): whether to use the ``pyhf.infer`` API or
            ``iminuit``, defaults to False (using ``pyhf.infer``)

    Raises:
        ValueError: if no POI is found

    Returns:
        RankingResults: fit results for parameters, and pre- and post-fit impacts
    """
    if fit_results is None:
        fit_results = _fit_model(
            model,
            data,
            init_pars=init_pars,
            fix_pars=fix_pars,
            par_bounds=par_bounds,
            strategy=strategy,
            maxiter=maxiter,
            tolerance=tolerance,
            custom_fit=custom_fit,
        )

    labels = model.config.par_names
    prefit_unc = model_utils.prefit_uncertainties(model)

    # use POI given by kwarg, fall back to POI specified in model
    poi_index = model_utils._poi_index(model, poi_name=poi_name)
    if poi_index is None:
        raise ValueError("no POI specified, cannot calculate ranking")

    nominal_poi = fit_results.bestfit[poi_index]

    # need to get values for parameter settings, as they will be partially changed
    # during the ranking (init/fix changes)
    # use parameter settings provided in function arguments if they exist, else defaults
    init_pars = init_pars or model.config.suggested_init()
    fix_pars = fix_pars or model.config.suggested_fixed()

    all_impacts = []
    for i_par, label in enumerate(labels):
        if i_par == poi_index:
            continue  # do not calculate impact of POI on itself
        log.info(f"calculating impact of {label} on {labels[poi_index]}")

        # hold current parameter constant
        fix_pars_ranking = fix_pars.copy()
        fix_pars_ranking[i_par] = True

        parameter_impacts = []
        # calculate impacts: pre-fit up, pre-fit down, post-fit up, post-fit down
        for np_val in [
            fit_results.bestfit[i_par] + prefit_unc[i_par],
            fit_results.bestfit[i_par] - prefit_unc[i_par],
            fit_results.bestfit[i_par] + fit_results.uncertainty[i_par],
            fit_results.bestfit[i_par] - fit_results.uncertainty[i_par],
        ]:
            # can skip pre-fit calculation for unconstrained parameters (their
            # pre-fit uncertainty is set to 0), and pre- and post-fit calculation
            # for fixed parameters (both uncertainties set to 0 as well)
            if np_val == fit_results.bestfit[i_par]:
                log.debug(f"impact of {label} is zero, skipping fit")
                parameter_impacts.append(0.0)
            else:
                init_pars_ranking = init_pars.copy()
                init_pars_ranking[i_par] = np_val  # value of current nuisance parameter
                fit_results_ranking = _fit_model(
                    model,
                    data,
                    init_pars=init_pars_ranking,
                    fix_pars=fix_pars_ranking,
                    par_bounds=par_bounds,
                    strategy=strategy,
                    maxiter=maxiter,
                    tolerance=tolerance,
                    custom_fit=custom_fit,
                )
                poi_val = fit_results_ranking.bestfit[poi_index]
                parameter_impact = poi_val - nominal_poi
                log.debug(
                    f"POI is {poi_val:.6f}, difference to nominal is "
                    f"{parameter_impact:.6f}"
                )
                parameter_impacts.append(parameter_impact)
        all_impacts.append(parameter_impacts)

    all_impacts_np = np.asarray(all_impacts)
    prefit_up = all_impacts_np[:, 0]
    prefit_down = all_impacts_np[:, 1]
    postfit_up = all_impacts_np[:, 2]
    postfit_down = all_impacts_np[:, 3]

    # remove parameter of interest from bestfit / uncertainty / labels
    # such that their entries match the entries of the impacts
    bestfit = np.delete(fit_results.bestfit, poi_index)
    uncertainty = np.delete(fit_results.uncertainty, poi_index)
    labels = np.delete(fit_results.labels, poi_index).tolist()

    ranking_results = RankingResults(
        bestfit, uncertainty, labels, prefit_up, prefit_down, postfit_up, postfit_down
    )
    return ranking_results


def scan(
    model: pyhf.pdf.Model,
    data: List[float],
    par_name: str,
    *,
    par_range: Optional[Tuple[float, float]] = None,
    n_steps: int = 11,
    init_pars: Optional[List[float]] = None,
    fix_pars: Optional[List[bool]] = None,
    par_bounds: Optional[List[Tuple[float, float]]] = None,
    strategy: Optional[Literal[0, 1, 2]] = None,
    maxiter: Optional[int] = None,
    tolerance: Optional[float] = None,
    custom_fit: bool = False,
) -> ScanResults:
    """Performs a likelihood scan over the specified parameter.

    If no parameter range is specified, center the scan around the best-fit result for
    the parameter that is being scanned, and scan over twice its uncertainty in each
    direction. The reported likelihood values are the differences between -2 log(L) at
    each point in the scan and the global minimum.

    Args:
        model (pyhf.pdf.Model): model to use in fits
        data (List[float]): data (including auxdata) the model is fit to
        par_name (str): name of parameter to scan over
        par_range (Optional[Tuple[float, float]], optional): upper and lower bounds of
            parameter in scan, defaults to None (automatically determine bounds)
        n_steps (int, optional): number of steps in scan, defaults to 10
        init_pars (Optional[List[float]], optional): list of initial parameter settings,
            defaults to None (use ``pyhf`` suggested inits)
        fix_pars (Optional[List[bool]], optional): list of booleans specifying which
            parameters are held constant, defaults to None (use ``pyhf`` suggestion)
        par_bounds (Optional[List[Tuple[float, float]]], optional): list of tuples with
            parameter bounds for fit, defaults to None (use ``pyhf`` suggested bounds)
        strategy (Optional[Literal[0, 1, 2]], optional): minimization strategy used by
            Minuit, can be 0/1/2, defaults to None (then uses ``pyhf`` default behavior
            of strategy 0 with user-provided gradients and 1 otherwise)
        maxiter (Optional[int], optional): allowed number of calls for minimization,
            defaults to None (use ``pyhf`` default of 100,000)
        tolerance (Optional[float]), optional): tolerance for convergence, for details
            see ``iminuit.Minuit.tol`` (uses EDM < 0.002*tolerance), defaults to
            None (use ``iminuit`` default of 0.1)
        custom_fit (bool, optional): whether to use the ``pyhf.infer`` API or
            ``iminuit``, defaults to False (using ``pyhf.infer``)

    Raises:
        ValueError: if parameter is not found in model

    Returns:
        ScanResults: includes parameter name, scanned values and 2*log(likelihood)
        offset
    """
    labels = model.config.par_names

    # get index of parameter with name par_name
    par_index = model_utils._parameter_index(par_name, labels)
    if par_index is None:
        raise ValueError(f"parameter {par_name} not found in model")

    # run a fit with the parameter not held constant, to find the best-fit point
    fit_results = _fit_model(
        model,
        data,
        init_pars=init_pars,
        fix_pars=fix_pars,
        par_bounds=par_bounds,
        strategy=strategy,
        maxiter=maxiter,
        tolerance=tolerance,
        custom_fit=custom_fit,
    )
    nominal_twice_nll = fit_results.best_twice_nll
    par_mle = fit_results.bestfit[par_index]
    par_unc = fit_results.uncertainty[par_index]

    if par_range is None:
        # if no parameter range is specified, use +/-2 sigma from the MLE
        par_range = (par_mle - 2 * par_unc, par_mle + 2 * par_unc)

    scan_values = np.linspace(par_range[0], par_range[1], n_steps)
    delta_nlls = np.zeros_like(scan_values)  # holds results

    # need to get values for parameter settings, as they will be partially changed
    # during the scan (init/fix changes)
    # use parameter settings provided in function arguments if they exist, else defaults
    init_pars = init_pars or model.config.suggested_init()
    fix_pars = fix_pars or model.config.suggested_fixed()

    fix_pars[par_index] = True  # hold scan parameter constant in fits

    log.info(
        f"performing likelihood scan for {par_name} in range ({par_range[0]:.3f}, "
        f"{par_range[1]:.3f}) with {n_steps} steps"
    )
    for i_par, par_value in enumerate(scan_values):
        log.debug(f"performing fit with {par_name} = {par_value:.3f}")
        init_pars_scan = init_pars.copy()
        init_pars_scan[par_index] = par_value
        scan_fit_results = _fit_model(
            model,
            data,
            init_pars=init_pars_scan,
            fix_pars=fix_pars,
            par_bounds=par_bounds,
            strategy=strategy,
            maxiter=maxiter,
            tolerance=tolerance,
            custom_fit=custom_fit,
        )
        # subtract best-fit
        delta_nlls[i_par] = scan_fit_results.best_twice_nll - nominal_twice_nll

    scan_results = ScanResults(par_name, par_mle, par_unc, scan_values, delta_nlls)
    return scan_results


def limit(
    model: pyhf.pdf.Model,
    data: List[float],
    *,
    bracket: Optional[Union[List[float], Tuple[float, float]]] = None,
    poi_tolerance: float = 0.01,
    maxsteps: int = 100,
    confidence_level: float = 0.95,
    poi_name: Optional[str] = None,
    init_pars: Optional[List[float]] = None,
    fix_pars: Optional[List[bool]] = None,
    par_bounds: Optional[List[Tuple[float, float]]] = None,
    strategy: Optional[Literal[0, 1, 2]] = None,
    maxiter: Optional[int] = None,
    tolerance: Optional[float] = None,
) -> LimitResults:
    """Calculates observed and expected upper parameter limits.

    Limits are calculated for the parameter of interest (POI) defined in the model.
    Brent's algorithm is used to automatically determine POI values to be tested. The
    desired confidence level can be configured, and defaults to 95%. In order to support
    setting the POI directly without model recompilation, this temporarily changes the
    POI in the model configuration.

    Args:
        model (pyhf.pdf.Model): model to use in fits
        data (List[float]): data (including auxdata) the model is fit to
        bracket (Optional[Union[List[float], Tuple[float, float]]], optional): the two
            POI values used to start the observed limit determination, the limit must
            lie between these values and the values must not be the same, defaults to
            None (then uses 0.1 as default lower value and the upper POI bound
            specified in the measurement as default upper value)
        poi_tolerance (float, optional): tolerance in POI value for convergence to
            target CLs value (1-``confidence_level``), defaults to 0.01
        maxsteps (int, optional): maximum number of steps for limit finding, defaults to
            100
        confidence_level (float, optional): confidence level for calculation, defaults
            to 0.95 (95%)
        poi_name (Optional[str], optional): limit is calculated for this parameter,
            defaults to None (use POI specified in workspace)
        init_pars (Optional[List[float]], optional): list of initial parameter settings,
            defaults to None (use ``pyhf`` suggested inits)
        fix_pars (Optional[List[bool]], optional): list of booleans specifying which
            parameters are held constant, defaults to None (use ``pyhf`` suggestion)
        par_bounds (Optional[List[Tuple[float, float]]], optional): list of tuples with
            parameter bounds for fit, defaults to None (use ``pyhf`` suggested bounds)
        strategy (Optional[Literal[0, 1, 2]], optional): minimization strategy used by
            Minuit, can be 0/1/2, defaults to None (then uses ``pyhf`` default behavior
            of strategy 0 with user-provided gradients and 1 otherwise)
        maxiter (Optional[int], optional): allowed number of calls for minimization,
            defaults to None (use ``pyhf`` default of 100,000)
        tolerance (Optional[float]), optional): tolerance for convergence, for details
            see ``iminuit.Minuit.tol`` (uses EDM < 0.002*tolerance), defaults to
            None (use ``iminuit`` default of 0.1)

    Raises:
        ValueError: if no POI is found
        ValueError: if lower and upper bracket value are the same
        ValueError: if starting brackets do not enclose the limit

    Returns:
        LimitResults: observed and expected limits, CLs values, and scanned points
    """
    _, initial_optimizer = pyhf.get_backend()  # store initial optimizer settings
    pyhf.set_backend(
        pyhf.tensorlib,
        pyhf.optimize.minuit_optimizer(
            verbose=1, strategy=strategy, maxiter=maxiter, tolerance=tolerance
        ),
    )

    # use POI given by kwarg, fall back to POI specified in model
    poi_index = model_utils._poi_index(model, poi_name=poi_name)
    if poi_index is None:
        raise ValueError("no POI specified, cannot calculate limit")

    # set POI name in model config to desired value, hypotest will pick this up
    # save original value to reset model later
    original_model_poi_name = model.config.poi_name
    model.config.set_poi(model.config.par_names[poi_index])

    # show two decimals only if confidence level in percent is not an integer
    cl_label = (
        f"{confidence_level:.{0 if (confidence_level * 100).is_integer() else 2}%}"
    )
    log.info(
        f"calculating {cl_label} confidence level upper limit for "
        f"{model.config.poi_name}"
    )

    # use par_bounds provided in function argument if they exist, else use default
    par_bounds = par_bounds or model.config.suggested_bounds()
    if par_bounds[model.config.poi_index][0] < 0:
        # set lower POI bound to zero (for use with qmu_tilde)
        par_bounds[model.config.poi_index] = (
            0.0,
            par_bounds[model.config.poi_index][1],
        )
        log.debug("setting lower parameter bound for POI to 0")

    # set default bracket to (0.1, upper POI bound in measurement) if needed
    bracket_left_default = 0.1
    bracket_right_default = par_bounds[model.config.poi_index][1]
    if bracket is None:
        bracket = (bracket_left_default, bracket_right_default)
    elif bracket[0] == bracket[1]:
        # set POI in model back to original value
        model.config.set_poi(original_model_poi_name)
        raise ValueError(f"the two bracket values must not be the same: {bracket}")

    cache_CLs: Dict[float, tuple] = {}  # cache storing all relevant results

    def _cls_minus_threshold(
        poi_val: float,
        model: pyhf.pdf.Model,
        data: List[float],
        cls_target: float,
        which_limit: int,
        limit_label: str,
    ) -> float:
        """Root of this function is the POI value at the CLs=``cls_target`` crossing.

        Returns 1-``cls_target`` for POI values below 0. Makes use of an external
        cache to avoid re-fitting known POI values and to store all relevant values.

        Args:
            poi_val (float): value for parameter of interest
            model (pyhf.pdf.Model): model to use in fits
            data (List[float]): data (including auxdata) the model is fit to
            cls_target (float): target CLs value to find by varying POI
            which_limit (int): which limit to run, 0: observed, 1: expected -2 sigma, 2:
                expected -1 sigma, 3: expected, 4: expected +1 sigma, 5: expected +2
                sigma
            limit_label (str): string to use when referring to the current limit

        Raises:
            ValueError: if root finding fails (usually due to starting bracket choice)

        Returns:
            float: absolute value of difference to CLs=``cls_target``
        """
        if poi_val <= 0:
            # no fit needed for negative POI value, return a default value
            log.debug(
                f"skipping fit for {model.config.poi_name} = {poi_val:.4f}, setting "
                "CLs = 1"
            )
            return 1 - cls_target  # distance of CLs = 1 to target CLs
        cache = cache_CLs.get(poi_val)
        if cache:
            observed, expected = cache  # use result from cache
        else:
            # calculate CLs
            results = pyhf.infer.hypotest(
                poi_val,
                data,
                model,
                init_pars=init_pars,
                fixed_params=fix_pars,
                par_bounds=par_bounds,
                test_stat="qtilde",
                return_expected_set=True,
            )
            observed = float(results[0])  # 1 value per scan point
            expected = np.asarray(results[1])  # 5 per point (with 1 and 2 sigma bands)
            cache_CLs.update({poi_val: (observed, expected)})
        current_CLs = np.hstack((observed, expected))[which_limit]
        log.debug(
            f"{model.config.poi_name} = {poi_val:.4f}, {limit_label} CLs = "
            f"{current_CLs:.4f}{' (cached)' if cache else ''}"
        )
        return current_CLs - cls_target

    # calculate all limits, one by one: observed, expected -2 sigma, expected -1 sigma,
    # expected, expected +1 sigma, expected +2 sigma
    cls_target = 1 - confidence_level
    limit_labels = [
        "observed",
        "expected -2 sigma",
        "expected -1 sigma",
        "expected",
        "expected +1 sigma",
        "expected +2 sigma",
    ]
    steps_total = 0
    all_limits = []
    all_converged = True
    for i_limit, limit_label in enumerate(limit_labels):
        log.info(f"determining {limit_label} upper limit")

        try:
            # find the 95% CL upper limit
            res = scipy.optimize.root_scalar(
                _cls_minus_threshold,
                bracket=bracket,
                args=(model, data, cls_target, i_limit, limit_label),
                method="brentq",
                options={"xtol": poi_tolerance, "maxiter": maxsteps},
            )
        except ValueError:
            # invalid starting bracket is most common issue
            log.error(
                f"CLs values at {bracket[0]:.4f} and {bracket[1]:.4f} do not bracket "
                f"CLs={cls_target:.4f}, try a different starting bracket"
            )
            # set POI in model back to original value
            model.config.set_poi(original_model_poi_name)
            raise

        if not res.converged:
            log.error(
                f"failed to converge after {res.function_calls} steps: {res.flag}"
            )
            all_converged = False
        else:
            log.info(f"successfully converged after {res.function_calls} steps")

        log.info(f"{limit_label} upper limit: {res.root:.4f}")
        all_limits.append(res.root)
        steps_total += res.function_calls

        # determine the starting bracket for the next limit calculation
        if i_limit < 5:
            # expected CLs values for next limit type that have been calculated already
            exp_CLs_next = np.asarray([exp[i_limit] for _, exp in cache_CLs.values()])
            # associated POI values
            poi_arr = np.fromiter(cache_CLs.keys(), dtype=float)

            # left: CLs has to be > cls_target, mask out values where CLs <= cls_target
            masked_CLs_left = np.where(exp_CLs_next <= cls_target, 1, exp_CLs_next)
            if sum(masked_CLs_left != 1) == 0:
                # all values are below cls_target, pick default lower bound
                bracket_left = bracket_left_default
            else:
                # find closest to CLs = cls_target from above
                bracket_left = poi_arr[np.argmin(masked_CLs_left)]

            # right: CLs has to be < cls_target, mask out values where CLs >= cls_target
            masked_CLs_right = np.where(exp_CLs_next >= cls_target, -1, exp_CLs_next)
            if sum(masked_CLs_right != -1) == 0:
                # all values are above cls_target, pick default upper bound
                bracket_right = bracket_right_default
            else:
                # find closest to CLs=cls_target from below
                bracket_right = poi_arr[np.argmax(masked_CLs_right)]

            bracket = (bracket_left, bracket_right)

    # set POI in model back to original values
    model.config.set_poi(original_model_poi_name)

    # report all results
    log.info(f"total of {steps_total} steps to calculate all limits")
    if not all_converged:
        log.error("one or more calculations did not converge, check log")
    log.info(f"summary of {cl_label} confidence level upper limits:")
    for i_limit, limit_label in enumerate(limit_labels):
        log.info(f"{limit_label.ljust(17)}: {all_limits[i_limit]:.4f}")

    # sort all CLs values and scanned POI points by increasing POI value
    poi_arr = np.fromiter(cache_CLs.keys(), dtype=float)
    sorted_indices = np.argsort(poi_arr)
    observed_CLs_np = np.asarray([obs for obs, _ in cache_CLs.values()])[sorted_indices]
    expected_CLs_np = np.asarray([exp for _, exp in cache_CLs.values()])[sorted_indices]
    poi_arr = poi_arr[sorted_indices]

    limit_results = LimitResults(
        all_limits[0],
        np.asarray(all_limits[1:]),
        observed_CLs_np,
        expected_CLs_np,
        poi_arr,
        confidence_level,
    )
    pyhf.set_backend(pyhf.tensorlib, initial_optimizer)  # restore optimizer settings
    return limit_results


def significance(
    model: pyhf.pdf.Model,
    data: List[float],
    *,
    poi_name: Optional[str] = None,
    init_pars: Optional[List[float]] = None,
    fix_pars: Optional[List[bool]] = None,
    par_bounds: Optional[List[Tuple[float, float]]] = None,
    strategy: Optional[Literal[0, 1, 2]] = None,
    maxiter: Optional[int] = None,
    tolerance: Optional[float] = None,
) -> SignificanceResults:
    """Calculates the discovery significance of a positive signal.

    Observed and expected p-values and significances are both calculated and reported.

    Args:
        model (pyhf.pdf.Model): model to use in fits
        data (List[float]): data (including auxdata) the model is fit to
        poi_name (Optional[str], optional): significance is calculated for this
            parameter, defaults to None (use POI specified in workspace)
        init_pars (Optional[List[float]], optional): list of initial parameter settings,
            defaults to None (use ``pyhf`` suggested inits)
        fix_pars (Optional[List[bool]], optional): list of booleans specifying which
            parameters are held constant, defaults to None (use ``pyhf`` suggestion)
        par_bounds (Optional[List[Tuple[float, float]]], optional): list of tuples with
            parameter bounds for fit, defaults to None (use ``pyhf`` suggested bounds)
        strategy (Optional[Literal[0, 1, 2]], optional): minimization strategy used by
            Minuit, can be 0/1/2, defaults to None (then uses ``pyhf`` default behavior
            of strategy 0 with user-provided gradients and 1 otherwise)
        maxiter (Optional[int], optional): allowed number of calls for minimization,
            defaults to None (use ``pyhf`` default of 100,000)
        tolerance (Optional[float]), optional): tolerance for convergence, for details
            see ``iminuit.Minuit.tol`` (uses EDM < 0.002*tolerance), defaults to
            None (use ``iminuit`` default of 0.1)

    Returns:
        SignificanceResults: observed and expected p-values and significances
    """
    _, initial_optimizer = pyhf.get_backend()  # store initial optimizer settings
    pyhf.set_backend(
        pyhf.tensorlib,
        pyhf.optimize.minuit_optimizer(
            verbose=1, strategy=strategy, maxiter=maxiter, tolerance=tolerance
        ),
    )

    # use POI given by kwarg, fall back to POI specified in model
    poi_index = model_utils._poi_index(model, poi_name=poi_name)
    if poi_index is None:
        raise ValueError("no POI specified, cannot calculate significance")

    # set POI name in model config to desired value, hypotest will pick this up
    # save original value to reset model later
    original_model_poi_name = model.config.poi_name
    model.config.set_poi(model.config.par_names[poi_index])

    log.info(f"calculating discovery significance for {model.config.poi_name}")
    obs_p_val, exp_p_val = pyhf.infer.hypotest(
        0.0,
        data,
        model,
        init_pars=init_pars,
        fixed_params=fix_pars,
        par_bounds=par_bounds,
        test_stat="q0",
        return_expected=True,
    )
    obs_p_val = float(obs_p_val)
    exp_p_val = float(exp_p_val)
    obs_significance = scipy.stats.norm.isf(obs_p_val, 0, 1)
    exp_significance = scipy.stats.norm.isf(exp_p_val, 0, 1)

    # set POI in model back to original values
    model.config.set_poi(original_model_poi_name)

    if obs_p_val >= 1e-3:
        log.info(f"observed p-value: {obs_p_val:.3%}")
    else:
        log.info(f"observed p-value: {obs_p_val:.3e}")
    log.info(f"observed significance: {obs_significance:.3f}")
    if exp_p_val >= 1e-3:
        log.info(f"expected p-value: {exp_p_val:.3%}")
    else:
        log.info(f"expected p-value: {exp_p_val:.3e}")
    log.info(f"expected significance: {exp_significance:.3f}")

    significance_results = SignificanceResults(
        obs_p_val, obs_significance, exp_p_val, exp_significance
    )
    pyhf.set_backend(pyhf.tensorlib, initial_optimizer)  # restore optimizer settings
    return significance_results
